#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

# pylint: disable=unicode-builtin

from __future__ import print_function
import os, sys, time, json
from glob import glob
if os.name == "posix":
    import fcntl


class _StateProp(object):
    """Defines a state property on a StateSource object."""

    def __init__(self, name, default, propType, validator):
        self.name = name
        self.default = default
        self.propType = propType
        self.validator = validator


class ExclusiveFile:
    """Wraps around file objects to ensure process has locked writes"""

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def __enter__(self):
        self._file = open(*self._args, **self._kwargs)
        # for now, only support locking *nix
        if os.name == "posix":
            num_retries = 10
            while True:
                try:
                    fcntl.flock(self._file.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                    break
                except OSError as timeout_exc:
                    num_retries -= 1
                    if num_retries < 0:
                        raise timeout_exc
                    time.sleep(.5)
        return self._file
        

    def __exit__(self, *args):
        self._file.flush()
        try:
            if os.name == "posix":
                fcntl.flock(self._file.fileno(), fcntl.LOCK_UN)
        finally:
            self._file.close()


class StateSource(object):
    """An object which has some savable application state."""

    def __init__(self, parent, name):
        self._parentStateSource = parent
        self._childStateSources = dict()
        self._stateSourceName = name
        self._stateSourceProperties = dict()

        # Register child source with the parent.
        if self._parentStateSource is not None:
            self._parentStateSource._registerChildStateSource(self)

    def _registerChildStateSource(self, child):
        """Registers a child StateSource with this source object."""
        self._childStateSources[child._stateSourceName] = child

    def GetChildStateSource(self, childName):
        """Returns the child StateSource corresponding to childName, or None"""
        return self._childStateSources.get(childName)

    def _getState(self):
        """Get this source's state dict from its parent source."""
        if self._parentStateSource is None:
            return dict()
        else:
            return self._parentStateSource._getChildState(self._stateSourceName)

    def _getChildState(self, childName):
        """Get a child source's state dict. This method guarantees that a dict
        will be return but does not guarantee anything about the contents of
        the dict.
        """
        state = self._getState()

        if childName in state:
            childState = state[childName]
            # Child state could be loaded from file as any JSON-serializable
            # type (int, str, etc.). Only return it if it is a dict. Otherwise,
            # fallback to empty dict.
            if isinstance(childState, dict):
                return childState

        # Create a new state dict for the child and save it in this source's
        # state dict.
        childState = dict()
        state[childName] = childState
        return childState

    def _typeCheck(self, value, prop):
        """Validate a value against a StateProp."""

        # Make sure the value has the correct type.
        valueType = type(value)
        if valueType is not prop.propType:
            str_types = [str]

            if valueType is int and prop.propType is float:
                pass # ints are valid for float types.
            elif prop.propType is str and valueType is str:
                pass # str can be used interchangeably.
            else:
                print("Value {} has type {} but state property {} has type {}.".format(
                        repr(value), valueType, repr(prop.name), prop.propType),
                        file=sys.stderr)
                print("    Using default value {}.".format(repr(prop.default)),
                        file=sys.stderr)
                return False

        # Make sure value passes custom validation. Otherwise, use default value.
        if prop.validator(value):
            return True
        else:
            print("Value {} did not pass custom validation for state property {}.".format(
                    repr(value), repr(prop.name)), file=sys.stderr)
            print("    Using default value {}.".format(repr(prop.default)),
                    file=sys.stderr)
            return False

    def _saveState(self):
        """Saves the source's state to the settings object's state buffer."""
        newState = dict()

        # Save state properties.
        self.onSaveState(newState)

        # Validate state properties.
        for name, value in tuple(newState.items()):
            if name not in self._stateSourceProperties:
                print("State property {} not defined. It will be removed.".format(
                        repr(name)), file=sys.stderr)
                del newState[name]
            prop = self._stateSourceProperties[name]
            if self._typeCheck(value, prop):
                newState[name] = value
            else:
                newState[name] = prop.default

        # Make sure no state properties were forgotten.
        for prop in self._stateSourceProperties.values():
            if prop.name not in newState:
                print("State property {} not saved.".format(repr(prop.name)),
                        file=sys.stderr)

        # Update the real state dict with the new state. This preserves unused
        # data loaded from the state file.
        self._getState().update(newState)

        # Save all child states.
        for child in self._childStateSources.values():
            child._saveState()

    def stateProperty(self, name, default, propType=None, validator=lambda value: True):
        """Validates and creates a new StateProp for this source. The property's
        value is returned so this method can be used during StateSource
        initialization."""

        # Make sure there are no conflicting state properties.
        if name in self._stateSourceProperties:
            raise RuntimeError("State property name {} already in use.".format(
                    repr(name)))

        # Grab state property type from default value if it was not defined.
        if propType is None:
            propType = type(default)

        # Make sure default value is valid.
        if not isinstance(default, propType):
            raise RuntimeError("Default value {} does not match type {}.".format(
                    repr(default), repr(propType)))
        if not validator(default):
            raise RuntimeError("Default value {} does not pass custom validation "
                    "for state property {}.".format(repr(default), repr(name)))

        prop = _StateProp(name, default, propType, validator)
        self._stateSourceProperties[name] = prop

        # Load the value from the state dict and validate it.
        state = self._getState()
        value = state.get(name, default)
        if self._typeCheck(value, prop):
            return value
        else:
            return prop.default

    def onSaveState(self, state):
        """Save the source's state properties to a dict."""
        raise NotImplementedError


class Settings(StateSource):
    """An object which encapsulates saving and loading of application state to
    a state file. When created, it loads state from a state file and stores it
    in a buffer. Its children sources can fetch their piece of state from the
    buffer. On save, this object tells its children to save their current
    states, then saves the buffer back to the state file.
    """

    def __init__(self, version, stateFilePath=None):
        # Settings should be the only StateSource with no parent or name.
        StateSource.__init__(self, None, None)
        self._version = version
        self._stateFilePath = stateFilePath
        self._versionsStateBuffer = None
        self._stateBuffer = None
        self._isEphemeral = (self._stateFilePath is None)

        self._loadState()

    def _loadState(self):
        """Loads and returns application state from a state file. If the file is
        not found, contains invalid JSON, does not contain a dictionary, an
        empty state is returned instead.
        """

        # Load the dict containing all versions of app state.
        if not self._isEphemeral:
            try:
                with open(self._stateFilePath, "r") as fp:
                    self._versionsStateBuffer = json.load(fp)
            except IOError as e:
                if os.path.isfile(self._stateFilePath):
                    print("Error opening state file: " + str(e), file=sys.stderr)
                else:
                    print("State file not found, a new one will be created.",
                            file=sys.stderr)
            except ValueError:
                print("State file contained invalid JSON. Please fix or delete " +
                        "it. Default settings will be used for this instance of " +
                        "USDView, but will not be saved.", file=sys.stderr)
                self._isEphemeral = True

        # Make sure JSON returned a dict.
        if not isinstance(self._versionsStateBuffer, dict):
            self._versionsStateBuffer = dict()

        # Load the correct version of the state dict.
        self._stateBuffer = self._versionsStateBuffer.get(self._version, None)
        if not isinstance(self._stateBuffer, dict):
            self._stateBuffer = dict()
            self._versionsStateBuffer[self._version] = self._stateBuffer

    # overrides StateSource._getState
    def _getState(self):
        """Gets the buffered state rather than asking its parent for its state.
        """
        return self._stateBuffer

    def save(self):
        """Inform all children to save their states, then write the state buffer
        back to the state file.
        """
        if not self._isEphemeral:
            self._saveState()
            try:
                with ExclusiveFile(self._stateFilePath, "w") as fp:
                    json.dump(self._versionsStateBuffer, fp,
                            indent=2, separators=(",", ": "))
            except IOError as e:
                print("Could not save state file: " + str(e), file=sys.stderr)

    def onSaveState(self, state):
        """Settings object has no state properties."""
        pass


class ConfigManager:
    """
    Class used to manage, read, and write the different saved settings that
    represent the usdview application's current state.
    """

    EXTENSION = "state.json"
    defaultConfig = ""

    def __init__(self, configDirPath):
        """Creates the manager instance.

        Parameters
        ----------
        configDirPath : str
            The directory that contains the state files
        """
        self.settings = None
        self._saveOnClose = False
        self._configDirPath = configDirPath
        self._configPaths = self._loadConfigPaths()

    def loadSettings(self, config, version, isEphemeral=False):
        """
        Loads the specified config. We wait to do this instead of loading in
        init to allow the manager to be created and read the list of available
        configs without actually doing the more expensive settings loading.

        Paramters
        ---------
        config : str
            The name of the config
        version : int
            Version number (used by the State class)
        isEphemeral : bool
            Usually when we use the default config we save all settings on app
            close (expected behavior of usdview before the advent of
            ConfigManager). If isEphemeral, we won't save no matter what
        """
        self._saveOnClose = (
            not isEphemeral and config == self.defaultConfig)
        self.settings = Settings(version, self._configPaths[config])

    def _loadConfigPaths(self):
        """Private method to load the config names and associated paths"""
        if not self._configDirPath:
            return {self.defaultConfig: None}
        # empty string used to include the trailing separator
        self._configDirPath = os.path.join(self._configDirPath, "")
        configPaths = {
            # len(EXTENSION) + 1 to get the '.' too
            p[len(self._configDirPath):-(len(self.EXTENSION) + 1)]: p
            for p in glob(os.path.join(
                self._configDirPath, "[a-z_]*." + self.EXTENSION))}
        configPaths[self.defaultConfig] = os.path.join(
            self._configDirPath, self.EXTENSION)
        return configPaths

    def getConfigs(self):
        """Gets the list of config names

        Returns
        -------
        list[str]
            List of all the avaiable config names in the _configDirPath
        """

        return sorted(self._configPaths.keys())

    def save(self, newName=None):
        """Saves the current state to the specified config

        Parameters
        ----------
        newName : str
            The name of the config we will be saving to (it may or may not
            exist in the _configDirPath). Iff same as defaultConfig, we save on
            application close.
        """
        if newName:
            self.settings._stateFilePath = os.path.join(
                self._configDirPath, newName + "." + self.EXTENSION)
            self._saveOnClose = (newName == self.defaultConfig)
        self.settings.save()

    def close(self):
        """Signal that application is closing"""
        if self._saveOnClose:
            self.settings.save()



#
# Copyright 2022 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

'''
Module that implements python version of HydraSceneDebugger.

NOTE: Source code for a C++ version of this widget is available (without open
      source build support) in pxr/imaging/hdui.

      That version (even wrapped in python as a whole) is strongly recommended
      as the basis for any application integration as it is higher performance.
'''

from .qt import QtCore, QtWidgets, QtGui
from pxr import Sdf
from ._usdviewq import (
    HydraObserver,
    DataSourceLocator,
    DataSourceLocatorSet,
    ContainerDataSource,
    VectorDataSource,
    SampledDataSource,
)

# -----------------------------------------------------------------------------

class HydraSceneDebugger(QtWidgets.QWidget):
    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self, parent)
        self.setWindowTitle("Hydra Scene Debugger")

        self.__mainLayout = QtWidgets.QVBoxLayout(self)
        self.__toolbarLayout = QtWidgets.QHBoxLayout()
        self.__mainLayout.addLayout(self.__toolbarLayout)
        self.__siChooser = _RegisteredSceneIndexChooser(self)
        self.__siChooser.sceneIndexSelected.connect(self.__SceneIndexSelected)

        self.__goToInputButton = QtWidgets.QPushButton("Inputs")
        self.__goToInputButtonMenu = QtWidgets.QMenu()
        self.__goToInputButton.setMenu(self.__goToInputButtonMenu)

        self.__goToInputButtonMenu.aboutToShow.connect(
            self.__FillGoToInputMenu)

        self.__nameLabel = QtWidgets.QLabel()

        self.__loggerButton = QtWidgets.QPushButton("Show Notice Logger")
        self.__loggerWidget = None

        self.__toolbarLayout.addWidget(self.__siChooser)
        self.__toolbarLayout.addWidget(self.__goToInputButton)
        self.__toolbarLayout.addWidget(self.__nameLabel, 10)
        self.__toolbarLayout.addStretch()
        self.__toolbarLayout.addWidget(self.__loggerButton)

        self.__observer = HydraObserver()

        self.__splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.__mainLayout.addWidget(self.__splitter, 10)

        self.__siTreeWidget = _SceneIndexTreeWidget(self.__observer)
        self.__splitter.addWidget(self.__siTreeWidget)

        self.__dsTreeWidget = _DataSourceTreeWidget(self)
        self.__splitter.addWidget(self.__dsTreeWidget)

        self.__valueTreeView = _DataSourceValueTreeView(self)
        self.__splitter.addWidget(self.__valueTreeView)

        self.__siTreeWidget.primSelected.connect(self.__PrimSelected)
        self.__dsTreeWidget.dataSourceSelected.connect(
            self.__DataSourceSelected)

        self.__siTreeWidget.primDirtied.connect(self.__PrimDirtied)

        self.__loggerButton.clicked.connect(self.__ShowLoggerWindow)

        self.__noticePollingTimer = QtCore.QTimer()
        self.__noticePollingTimer.setInterval(16)
        self.__noticePollingTimer.timeout.connect(self.__PollForNotices)
        self.__noticePollingTimer.start()

    def __PollForNotices(self):
        if not self.__observer.HasPendingNotices():
            return

        notices = self.__observer.GetPendingNotices()
        if self.__loggerWidget:
            self.__loggerWidget.AddNotices(notices)

        for noticeType, entries in notices:
            if noticeType == 'added':
                self.__siTreeWidget.PrimsAdded(entries)
            elif noticeType == 'removed':
                self.__siTreeWidget.PrimsRemoved(entries)
            elif noticeType == 'dirtied':
                self.__siTreeWidget.PrimsDirtied(entries)

    def __SceneIndexSelected(self, name):
        self.__observer.TargetToNamedSceneIndex(name)
        self.__UpdateNameLabel(name)
        self.__SceneIndexChanged()

    def __SceneIndexChanged(self):
        # XXX for initial testing, clear and build /root
        self.__siTreeWidget.Requery(self.__observer)

    def __PrimSelected(self, primPath, dataSourceProxy):
        self.__valueTreeView.SetDataSource(None)
        self.__dsTreeWidget.SetPrimDataSource(primPath, dataSourceProxy)

    def __PrimDirtied(self, primPath, locators):
        primType, dataSource = self.__observer.GetPrim(primPath)
        self.__dsTreeWidget.PrimDirtied(primPath, dataSource, locators)

    def __DataSourceSelected(self, dataSourceProxy):
        self.__valueTreeView.SetDataSource(dataSourceProxy.value)

    def __FillGoToInputMenu(self):
        menu = self.__goToInputButtonMenu;
        menu.clear()
        menuTreeWidget = QtWidgets.QTreeWidget()
        menuTreeWidget.setHeaderHidden(True);
        menuTreeWidget.setAllColumnsShowFocus(True);
        menuTreeWidget.setMouseTracking(True);
        menuTreeWidget.setSizeAdjustPolicy(
            QtWidgets.QAbstractScrollArea.SizeAdjustPolicy.\
                AdjustToContentsOnFirstShow)

        def _ItemEntered(item):
            menuTreeWidget.setCurrentItem(item, 0,
                QtCore.QItemSelectionModel.SelectionFlag.Select |
                    QtCore.QItemSelectionModel.SelectionFlag.Clear)

        menuTreeWidget.itemEntered.connect(_ItemEntered)

        def _ItemClicked(item, column):
            if hasattr(item, 'indices'):
                self.__observer.TargetToInputSceneIndex(item.indices)
                self.__UpdateNameLabel()
                self.__SceneIndexChanged()
            menu.close()

        menuTreeWidget.itemClicked.connect(_ItemClicked)

        queue = [([], menuTreeWidget)]

        while queue:
            indices, itemParent = queue.pop()
            names = self.__observer.GetInputDisplayNames(indices)
            for i, name in enumerate(names):
               item = QtWidgets.QTreeWidgetItem(itemParent, [name])
               item.indices = indices + [i]
               item.setExpanded(True)
               queue.append((item.indices, item))

        widgetAction = QtWidgets.QWidgetAction(menu)
        widgetAction.setDefaultWidget(menuTreeWidget)

        menu.addAction(widgetAction)

    def __UpdateNameLabel(self, prefix=None):
        value = "<b><i>( %s )</i></b> %s" % (self.__observer.GetDisplayName(),
            '' if prefix is None else prefix)
        self.__nameLabel.setText(value)

        if self.__loggerWidget:
            self.__loggerWidget.SetText(value)

        self.__goToInputButton.setEnabled(
            bool(self.__observer.GetInputDisplayNames([])))

    def __ShowLoggerWindow(self):
        if not self.__loggerWidget:
            self.__loggerWidget = _NoticeLoggingWidget(None)
            self.__loggerWidget.SetText(self.__nameLabel.text())

        self.__loggerWidget.show()


# -----------------------------------------------------------------------------

class _RegisteredSceneIndexChooser(QtWidgets.QPushButton):
    sceneIndexSelected = QtCore.Signal(str)

    def __init__(self, parent=None):
        QtWidgets.QPushButton.__init__(self, "Choose Scene Index", parent)

        self.__menu = QtWidgets.QMenu()
        self.setMenu(self.__menu)


        self.__menu.aboutToShow.connect(self.__AboutToShow)
        self.__menu.triggered.connect(self.__Triggered)

    def __AboutToShow(self):
        self.__menu.clear()
        names = HydraObserver.GetRegisteredSceneIndexNames()
        if names:
            for name in names:
                self.__menu.addAction(name)
        else:
            self.__menu.addAction("No Registered Names").setEnabled(False)

    def __Triggered(self, action):
        self.sceneIndexSelected.emit(str(action.text()))

# -----------------------------------------------------------------------------

# holds a container data source or None
class _DataSourceProxy:
    def __init__(self, value=None):
        self.value = value

class _SceneIndexTreeWidget(QtWidgets.QTreeWidget):
    primSelected = QtCore.Signal(Sdf.Path, _DataSourceProxy)
    primDirtied = QtCore.Signal(Sdf.Path, DataSourceLocatorSet)

    def __init__(self, observer, parent=None):
        QtWidgets.QTreeWidget.__init__(self, parent)
        self.__observer = observer
        self.__expandedSet = set()
        self.__primItems = {}

        self.setHeaderLabels(["Name", "Type"])

        header = self.header()
        header.setSectionResizeMode(0, header.ResizeMode.Stretch);
        header.setSectionResizeMode(1, header.ResizeMode.Fixed);
        header.resizeSection(1, self.fontMetrics().averageCharWidth() * 10);
        header.setStretchLastSection(False);

        self.itemSelectionChanged.connect(self.__ItemSelectionChanged)
        self.itemExpanded.connect(self.__ItemExpanded)
        self.itemCollapsed.connect(self.__ItemCollapsed)

    def contextMenuEvent(self, event):
        item = self.itemAt(event.pos())

        if item:
            globalPos = QtCore.QPoint(event.pos().x(),
                self.visualItemRect(item).bottom())

            if self.header().isVisible():
                globalPos = QtCore.QPoint(
                    globalPos.x(), globalPos.y() + self.header().height())

            if isinstance(item, _SceneIndexPrimTreeWidgetItem):
                menu = QtWidgets.QMenu()
                menu.addAction("type: " + str(item.text(1))).setEnabled(False)
                menu.addSeparator()
                menu.exec_(self.mapToGlobal(globalPos))

    def PrimsAdded(self, entries):
        for primPath, primType in entries:
            item = self._GetPrimItem(primPath)
            item.setText(1, primType)
            if item.isSelected():
                self.itemSelectionChanged.emit()

    def PrimsRemoved(self, entries):
        sortState = self.isSortingEnabled()

        try:
            for primPath in entries:
                item = self._GetPrimItem(primPath, False)
                if item:
                    parentItem = item.parent()
                    if parentItem is not None:
                        parentItem.takeChild(parentItem.indexOfChild(item))
                    
                    # XXX Items are currently stored flat so this loop will not
                    #     scale for lots of items
                    self._RemoveSubtree(primPath)

        finally:
            self.setSortingEnabled(sortState)

    def PrimsDirtied(self, entries):
        items = self.selectedItems()
        if not items:
            return

        item = items[0]
        if not isinstance(item, _SceneIndexPrimTreeWidgetItem):
            return

        selectedPath = item.GetPrimPath()

        selectedItemLocators = DataSourceLocatorSet()
        for primPath, dirtiedLocators in entries:
            if primPath == selectedPath:
                selectedItemLocators.insert(dirtiedLocators)

        self.primDirtied.emit(selectedPath, selectedItemLocators)

    def _GetPrimItem(self, primPath, createIfNecessary=True):
        if primPath in self.__primItems:
            return self.__primItems[primPath]

        if not createIfNecessary:
            return None

        if primPath == Sdf.Path.absoluteRootPath:
            parentItem = self.invisibleRootItem()
        else:
            parentItem = self._GetPrimItem(primPath.GetParentPath(), True)

        if parentItem is None:
            return None

        item = _SceneIndexPrimTreeWidgetItem(parentItem, primPath)
        self.__primItems[primPath] = item
        return item

    def _AddPrimItem(self, primPath, item):
        self.__primItems[primPath] = item

    def _RemoveSubtree(self, primPath):
        item = self._GetPrimItem(primPath, createIfNecessary=False)
        if item is None:
            return

        itemParent = item.parent()
        if itemParent is not None:
            itemParent.takeChild(itemParent.indexOfChild(item))

        self._RemoveSubtreeEntries(primPath)

    def _RemoveSubtreeEntries(self, primPath):
        toRemove = []
        for itemPrimPath in self.__primItems:
            if itemPrimPath.HasPrefix(primPath):
                toRemove.append(primPath)

        for p in toRemove:
            if p in self.__primItems:
                del self.__primItems[p]

    def Requery(self, observer, lazy=True):
        self.clear()
        self.__primItems.clear()
        self.__primItems[Sdf.Path.absoluteRootPath] = (
            _SceneIndexPrimTreeWidgetItem(self.invisibleRootItem(),
                Sdf.Path.absoluteRootPath, True))

    def PrimIsExpanded(self, primPath):
        return primPath in self.__expandedSet

    def AddToExpandedSet(self, primPath):
        self.__expandedSet.add(primPath)

    def RemoveFromExpandedSet(self, primPath):
        self.__expandedSet.discard(primPath)

    def __ItemSelectionChanged(self):
        items = self.selectedItems()
        if not items:
            self.primSelected.emit(Sdf.Path(), _DataSourceProxy())
            return

        item = items[0]
        if isinstance(item, _SceneIndexPrimTreeWidgetItem):
            primPath = item.GetPrimPath()
            primType, dataSource = self.__observer.GetPrim(primPath)
            self.primSelected.emit(item.GetPrimPath(),
                _DataSourceProxy(dataSource))

    def __ItemExpanded(self, item):
        if isinstance(item, _SceneIndexPrimTreeWidgetItem):
            item.WasExpanded(self, self.__observer)

    def __ItemCollapsed(self, item):
        if isinstance(item, _SceneIndexPrimTreeWidgetItem):
            self.RemoveFromExpandedSet(item.GetPrimPath())

# -----------------------------------------------------------------------------

class _SceneIndexPrimTreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, parent, primPath, queryOnExpansion=False):
        QtWidgets.QTreeWidgetItem.__init__(self, parent)
        self.__primPath = primPath
        self.__queryOnExpansion = queryOnExpansion

        if queryOnExpansion:
            self.setChildIndicatorPolicy(
                self.ChildIndicatorPolicy.ShowIndicator)

        if primPath.IsPropertyPath():
            name = "." + primPath.name
        else:
            name = primPath.name

        self.setText(0, name)

        if self.treeWidget().PrimIsExpanded(primPath):
            self.setExpanded(True)

    def GetPrimPath(self):
        return self.__primPath

    def WasExpanded(self, treeWidget, observer):
        treeWidget.AddToExpandedSet(self.__primPath)

        if not self.__queryOnExpansion:
            return

        self.__queryOnExpansion = False

        count = self.childCount()
        if count:
            for i in range(0, count):
                childItem = self.child(i)
                if isinstance(childItem, _SceneIndexPrimTreeWidgetItem):
                    treeWidget._RemoveSubtree(childItem)

        for childPath in observer.GetChildPrimPaths(self.GetPrimPath()):
            childPrimType, childPrimDataSource = observer.GetPrim(childPath)
            childItem = _SceneIndexPrimTreeWidgetItem(self, childPath, True)
            childItem.setText(1, childPrimType)
            treeWidget._AddPrimItem(childPath, childItem)

        if not self.childCount():
            self.setChildIndicatorPolicy(
                self.ChildIndicatorPolicy.DontShowIndicator)

# -----------------------------------------------------------------------------

class _DataSourceTreeWidget(QtWidgets.QTreeWidget):
    dataSourceSelected = QtCore.Signal(_DataSourceProxy)

    def __init__(self, parent):
        QtWidgets.QTreeWidget.__init__(self, parent)

        self.setHeaderLabels(['Name'])
        self.setAllColumnsShowFocus(True)
        self.__expandedSet = set()

        self.itemExpanded.connect(self.__ItemExpanded)
        self.itemCollapsed.connect(self.__ItemCollapsed)
        self.itemSelectionChanged.connect(self.__ItemSelectionChanged)

    def SetPrimDataSource(self, primPath, dataSource):
        self.clear()
        if dataSource.value is not None:
            item = _DataSourceTreeWidgetItem(
                DataSourceLocator(),
                self.invisibleRootItem(),
                dataSource.value)
            item.setText(0, '<%s>' % (primPath.name,))

    def PrimDirtied(self, primPath, dataSource, locatorSet):
        # loop over existing items to determine which require data source
        # updates

        taskQueue = [self.topLevelItem(0)]

        while taskQueue:
            item = taskQueue[-1]
            del taskQueue[-1]

            if not isinstance(item, _DataSourceTreeWidgetItem):
                continue

            loc = item.GetLocator()

            addChildren = False

            if not loc.IsEmpty():
                if locatorSet.Contains(loc):
                    # dirty here, we'll need a new data source
                    # no need to add children as SetDirty will handle that
                    item.SetDirty(dataSource.Get(loc))
                elif locatorSet.Intersects(loc):
                    addChildren = True
            else:
                addChildren = True

            if addChildren:
                for i in range(item.childCount()):
                    taskQueue.append(item.child(i))


        # Force a selection change on the current item so that the value column
        # re-pulls on the data source
        items = self.selectedItems()
        if items:
            item = items[0]
            if isinstance(item, _DataSourceTreeWidgetItem):
                if locatorSet.Intersects(item.GetLocator()):
                    self.itemSelectionChanged.emit()


    def _GetExpandedSet(self):
        return self.__expandedSet

    def __ItemExpanded(self, item):
        if isinstance(item, _DataSourceTreeWidgetItem):
            item.WasExpanded()

    def __ItemCollapsed(self, item):
        if isinstance(item, _DataSourceTreeWidgetItem):
            item.WasCollapsed()
    
    def __ItemSelectionChanged(self):
        items = self.selectedItems()
        if not items:
            return

        item = items[0]
        if not isinstance(item, _DataSourceTreeWidgetItem):
            return

        self.dataSourceSelected.emit(_DataSourceProxy(item.GetDataSource()))

# -----------------------------------------------------------------------------

class _DataSourceTreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, locator, parentItem, dataSource):
        QtWidgets.QTreeWidgetItem.__init__(self, parentItem)
        self.__locator = locator
        self.__dataSource = dataSource
        self.__childrenBuilt = False

        if not locator.IsEmpty():
            self.setText(0, locator.GetLastElement())

        if isinstance(self.__dataSource, (ContainerDataSource,
                VectorDataSource,)):
            self.setChildIndicatorPolicy(
                self.ChildIndicatorPolicy.ShowIndicator)
        else:
            self.setChildIndicatorPolicy(
                self.ChildIndicatorPolicy.DontShowIndicator)

        if self.__IsInExpandedSet():
            #TODO, determine whether this is necessary to defer anymore
            self.setExpanded(True)

    def __IsInExpandedSet(self):
        return self.__locator in self.treeWidget()._GetExpandedSet()

    def __SetIsInExpandedSet(self, state):
        if state:
            self.treeWidget()._GetExpandedSet().add(self.__locator)
        else:
            self.treeWidget()._GetExpandedSet().discard(self.__locator)

    def WasExpanded(self):
        self.__SetIsInExpandedSet(True)

        if self.__childrenBuilt:
            return

        self.__BuildChildren()

    def WasCollapsed(self):
        self.__SetIsInExpandedSet(False)

    def SetDirty(self, dataSource):
        if self.__childrenBuilt:
            if isinstance(dataSource, ContainerDataSource):
                # confirm that existing data source is also a container
                # if not, rebuild entirely
                if not isinstance(self.__dataSource, ContainerDataSource):
                    self.__dataSource = dataSource
                    self.__RebuildChildren()
                    return

                usedNames = set()
                itemsToRemove = []

                for i in range(self.childCount()):
                    childItem = self.child(i)
                    if not isinstance(childItem, _DataSourceTreeWidgetItem):
                        continue

                    childName = childItem.GetLocator().GetLastElement()
                    childDs = dataSource.Get(childName)
                    usedNames.add(childName)

                    if childDs is not None:
                        childItem.SetDirty(childDs)
                    else:
                        itemsToRemove.append(childItem)

                # add any new items
                for childName in dataSource.GetNames():
                    if childName not in usedNames:
                        childDs = dataSource.Get(childName)
                        if childDs is not None:
                            _DataSourceTreeWidgetItem(
                                self.__locator.Append(childName),
                                self,
                                childDs)

                for item in itemsToRemove:
                    self.takeChild(self.indexOfChild(item))
            elif isinstance(dataSource, VectorDataSource):

                # confirm that existing data source is also a vector
                # of the same length (could reuse items but probably not
                # worth the extra complexity)
                if (not isinstance(self.__dataSource, VectorDataSource) or
                        self.childCount() != dataSource.GetNumElements()):
                    self.__dataSource = dataSource
                    self.__RebuildChildren()
                    return


                for i in range(dataSource.GetNumElements()):
                    childItem = self.child(i)
                    if isinstance(childItem, _DataSourceTreeWidgetItem):
                        childItem.SetDirty(dataSource.GetElement(i))

        self.__dataSource = dataSource

    def GetDataSource(self):
        return self.__dataSource

    def GetLocator(self):
        return self.__locator

    def __RebuildChildren(self):
        self.takeChildren()
        self.__BuildChildren()
        
    def __BuildChildren(self):
        self.__childrenBuilt = True
        
        if isinstance(self.__dataSource, ContainerDataSource):
            usedNames = set()
            for childName in self.__dataSource.GetNames():
                if childName in usedNames:
                    continue
                usedNames.add(childName)
                childDataSource = self.__dataSource.Get(childName)
                if childDataSource is not None:
                    _DataSourceTreeWidgetItem(
                        self.__locator.Append(childName),
                        self,
                        childDataSource)

        elif isinstance(self.__dataSource, VectorDataSource):
            for i in range(self.__dataSource.GetNumElements()):
                childName = 'i%i' % (i,)
                _DataSourceTreeWidgetItem(
                    self.__locator.Append(childName),
                    self,
                    self.__dataSource.GetElement(i))

# -----------------------------------------------------------------------------

class _ValueItemModel(QtCore.QAbstractItemModel):
    def __init__(self, value, typeStr, parent=None):
        QtCore.QAbstractItemModel.__init__(self, parent)
        self.__value = value
        self.__isArray = hasattr(value, 'AsList')
        self.__typeStr = typeStr

    # base is good for scalars we'll use str() on the held type
    def data(self, modelIndex, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if role != QtCore.Qt.ItemDataRole.DisplayRole:
            return None

        if self.__isArray:
            if  modelIndex.row() <= len(self.__value):
                if modelIndex.column() == 0:
                    return str(self.__value[modelIndex.row()])
                else:
                    return modelIndex.row()
        else:
            if modelIndex.row() == 0 and modelIndex.column() == 0:
                return str(self.__value)

        return None

    def headerData(self, section, orientation, role=QtCore.Qt.ItemDataRole.DisplayRole):

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if section == 0:
                return self.__typeStr
            elif section == 1 and self.__isArray:
                return "Index"

        return None

    def parent(self, model):
        return QtCore.QModelIndex()

    def columnCount(self, parent=QtCore.QModelIndex()):
        if self.__isArray:
            return 2
        return 1

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid() or parent.column() > 0:
            return 0

        if self.__isArray:
            return len(self.__value)

        return 1

    def index(self, row, column, parent=None):
        return self.createIndex(row, column)

# -----------------------------------------------------------------------------

class _DataSourceValueTreeView(QtWidgets.QTreeView):
    def __init__(self, parent):
        QtWidgets.QTreeView.__init__(self, parent)
        self.__dataSource = None
        self.setUniformRowHeights(True)
        self.setItemsExpandable(False)


    def SetDataSource(self, dataSource):
        self.__dataSource = dataSource
        if isinstance(dataSource, SampledDataSource):

            # NOTE: GetTypeString internally has to call GetValue so this
            #       is happening twice. If that's a performance issue ever,
            #       we can add GetValueAndTypeString to get both at once.
            self.setModel(_ValueItemModel(dataSource.GetValue(0.0),
                dataSource.GetTypeString(), self))

            header = self.header()
            header.setSectionResizeMode(0, header.ResizeMode.Stretch)
            if header.count() > 1:
                header.setSectionResizeMode(1, header.ResizeMode.Fixed)
                header.resizeSection(1,
                    self.fontMetrics().averageCharWidth() * 10)
                header.setStretchLastSection(False)
            else:
                header.setStretchLastSection(True)
        else:
            self.setModel(None)

    def Refresh(self):
        self.SetDataSource(self.__dataSource)

# -----------------------------------------------------------------------------

class _NoticeLoggingWidget(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)

        self.__mainLayout = QtWidgets.QVBoxLayout(self)
        self.__toolbarLayout = QtWidgets.QHBoxLayout()
        self.__mainLayout.addLayout(self.__toolbarLayout)

        self.__startStopButton = QtWidgets.QPushButton("Start Recording")
        self.__clearButton = QtWidgets.QPushButton("Clear")

        self.__label = QtWidgets.QLabel('', self)

        self.__toolbarLayout.addWidget(self.__startStopButton)
        self.__toolbarLayout.addWidget(self.__label, 10)
        self.__toolbarLayout.addStretch()
        self.__toolbarLayout.addWidget(self.__clearButton)

        self.__recording = False

        self.__treeView = _ObserverLoggingTreeView(self)
        self.__mainLayout.addWidget(self.__treeView)


        self.__clearButton.clicked.connect(self.__Clear)

        self.__startStopButton.clicked.connect(self.__ToggleRecordingState)

    def AddNotices(self, entries):
        if not self.__recording:
            return
        self.__treeView.GetModel().AddEntries(entries)

    def SetText(self, value):
        self.__label.setText(value)

    def __Clear(self):
        self.__treeView.GetModel().Clear()

    def __ToggleRecordingState(self):
        self.__recording = not self.__recording
        self.__startStopButton.setText(
            "Start Recording" if not self.__recording else "Stop Recording")


# -----------------------------------------------------------------------------

class _ObserverLoggingTreeView(QtWidgets.QTreeView):
    def __init__(self, parent=None):
        QtWidgets.QTreeView.__init__(self, parent)
        self.__model = _ObserverLoggingItemModel(self)
        self.setMinimumWidth(512)
        self.setUniformRowHeights(True)
        self.setModel(self.__model)
        self.header().resizeSection(0, 384)

    def GetModel(self):
        return self.__model

# -----------------------------------------------------------------------------


class _ObserverLoggingItemModel(QtCore.QAbstractItemModel):

    def __init__(self, parent):
        QtCore.QAbstractItemModel.__init__(self, parent)
        self.__entries = []

    def AddEntries(self, newEntries):
        if not newEntries:
            return

        beforeLength = len(self.__entries)
        self.beginInsertRows(QtCore.QModelIndex(), beforeLength,
            beforeLength + len(newEntries) - 1)
        
        for i, newEntry in enumerate(newEntries):
            # stow the index into the entry in service of modelIndex
            # "internalPointer"
            self.__entries.append((i + beforeLength, newEntry))

        self.endInsertRows()

    def Clear(self):
        self.beginResetModel()
        self.__entries.clear()
        self.endResetModel()

    def data(self, modelIndex, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if role != QtCore.Qt.ItemDataRole.DisplayRole:
            return None

        internalPtr = modelIndex.internalPointer()
        if internalPtr:
            parentIndex, (entryType, entryValues) = internalPtr
            entryValue = entryValues[modelIndex.row()]

            if modelIndex.column() == 0:
                if entryType == 'removed':
                    return str(entryValue)
                return str(entryValue[0])

            elif modelIndex.column() == 1:
                if entryType == 'dirtied':
                    return entryValue[1].AsString()
                elif entryType == 'added':
                    return str(entryValue[1])

        else:
            if modelIndex.column() == 0:
                # return the notice type string
                return self.__entries[modelIndex.row()][1][0]

        return None

    def headerData(self, section, orientation, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if section == 0:
                return "Notice Type/ Prim Path"
            elif section == 1:
                return "Value"

        return None

    def index(self, row, column, parent=None):
        if parent.isValid():
            # safeguard against child of child

            if parent.internalPointer():
                return QtCore.QModelIndex()

            return self.createIndex(row, column, self.__entries[parent.row()])

        if row >= 0 and row < len(self.__entries):
            return self.createIndex(row, column, None)

        return QtCore.QModelIndex()

    def parent(self, modelIndex):
        internalPtr = modelIndex.internalPointer()
        # top-level items have no pointer so won't have a parent
        if not internalPtr:
            return QtCore.QModelIndex()

        return self.createIndex(internalPtr[0], modelIndex.row(), None)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 2

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.column() > 0:
            return 0

        if parent.isValid():
            internalPtr = parent.internalPointer()
            if internalPtr:
                return 0
            else:
                return len(self.__entries[parent.row()][1][1])

        return len(self.__entries)







# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'primLegendUI.ui'
##
## Created by: Qt User Interface Compiler version 6.11.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGraphicsView, QGridLayout, QLabel,
    QSizePolicy, QVBoxLayout, QWidget)

class Ui_PrimLegend(object):
    def setupUi(self, PrimLegend):
        if not PrimLegend.objectName():
            PrimLegend.setObjectName(u"PrimLegend")
        PrimLegend.resize(438, 131)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(PrimLegend.sizePolicy().hasHeightForWidth())
        PrimLegend.setSizePolicy(sizePolicy)
        self.primLegendLayoutContainer = QVBoxLayout(PrimLegend)
        self.primLegendLayoutContainer.setObjectName(u"primLegendLayoutContainer")
        self.primLegendLayout = QGridLayout()
        self.primLegendLayout.setObjectName(u"primLegendLayout")
        self.primLegendColorHasArcs = QGraphicsView(PrimLegend)
        self.primLegendColorHasArcs.setObjectName(u"primLegendColorHasArcs")
        self.primLegendColorHasArcs.setMaximumSize(QSize(20, 15))

        self.primLegendLayout.addWidget(self.primLegendColorHasArcs, 0, 0, 1, 1)

        self.primLegendLabelHasArcs = QLabel(PrimLegend)
        self.primLegendLabelHasArcs.setObjectName(u"primLegendLabelHasArcs")
        font = QFont()
        font.setBold(False)
        font.setItalic(False)
        self.primLegendLabelHasArcs.setFont(font)

        self.primLegendLayout.addWidget(self.primLegendLabelHasArcs, 0, 1, 1, 1)

        self.primLegendColorInstance = QGraphicsView(PrimLegend)
        self.primLegendColorInstance.setObjectName(u"primLegendColorInstance")
        self.primLegendColorInstance.setMaximumSize(QSize(20, 15))

        self.primLegendLayout.addWidget(self.primLegendColorInstance, 0, 2, 1, 1)

        self.primLegendLabelInstance = QLabel(PrimLegend)
        self.primLegendLabelInstance.setObjectName(u"primLegendLabelInstance")
        self.primLegendLabelInstance.setFont(font)

        self.primLegendLayout.addWidget(self.primLegendLabelInstance, 0, 3, 1, 1)

        self.primLegendColorPrototype = QGraphicsView(PrimLegend)
        self.primLegendColorPrototype.setObjectName(u"primLegendColorPrototype")
        self.primLegendColorPrototype.setMaximumSize(QSize(20, 15))

        self.primLegendLayout.addWidget(self.primLegendColorPrototype, 0, 4, 1, 1)

        self.primLegendLabelPrototype = QLabel(PrimLegend)
        self.primLegendLabelPrototype.setObjectName(u"primLegendLabelPrototype")
        self.primLegendLabelPrototype.setFont(font)

        self.primLegendLayout.addWidget(self.primLegendLabelPrototype, 0, 5, 1, 1)

        self.primLegendColorNormal = QGraphicsView(PrimLegend)
        self.primLegendColorNormal.setObjectName(u"primLegendColorNormal")
        self.primLegendColorNormal.setMaximumSize(QSize(20, 15))

        self.primLegendLayout.addWidget(self.primLegendColorNormal, 0, 6, 1, 1)

        self.primLegendLabelNormal = QLabel(PrimLegend)
        self.primLegendLabelNormal.setObjectName(u"primLegendLabelNormal")
        self.primLegendLabelNormal.setFont(font)

        self.primLegendLayout.addWidget(self.primLegendLabelNormal, 0, 7, 1, 1)


        self.primLegendLayoutContainer.addLayout(self.primLegendLayout)

        self.primLegendLabelContainer = QVBoxLayout()
        self.primLegendLabelContainer.setObjectName(u"primLegendLabelContainer")
        self.primLegendLabelDimmed = QLabel(PrimLegend)
        self.primLegendLabelDimmed.setObjectName(u"primLegendLabelDimmed")

        self.primLegendLabelContainer.addWidget(self.primLegendLabelDimmed)

        self.primLegendLabelFontsAbstract = QLabel(PrimLegend)
        self.primLegendLabelFontsAbstract.setObjectName(u"primLegendLabelFontsAbstract")

        self.primLegendLabelContainer.addWidget(self.primLegendLabelFontsAbstract)

        self.primLegendLabelFontsUndefined = QLabel(PrimLegend)
        self.primLegendLabelFontsUndefined.setObjectName(u"primLegendLabelFontsUndefined")

        self.primLegendLabelContainer.addWidget(self.primLegendLabelFontsUndefined)

        self.primLegendLabelFontsDefined = QLabel(PrimLegend)
        self.primLegendLabelFontsDefined.setObjectName(u"primLegendLabelFontsDefined")

        self.primLegendLabelContainer.addWidget(self.primLegendLabelFontsDefined)


        self.primLegendLayoutContainer.addLayout(self.primLegendLabelContainer)


        self.retranslateUi(PrimLegend)

        QMetaObject.connectSlotsByName(PrimLegend)
    # setupUi

    def retranslateUi(self, PrimLegend):
        PrimLegend.setProperty(u"comment", QCoreApplication.translate("PrimLegend", u"\n"
"     Copyright 2017 Pixar                                                                   \n"
"                                                                                            \n"
"     Licensed under the terms set forth in the LICENSE.txt file available at\n"
"     https://openusd.org/license.\n"
"  ", None))
        self.primLegendLabelHasArcs.setText(QCoreApplication.translate("PrimLegend", u"HasArcs", None))
        self.primLegendLabelInstance.setText(QCoreApplication.translate("PrimLegend", u"Instance", None))
        self.primLegendLabelPrototype.setText(QCoreApplication.translate("PrimLegend", u"Prototype", None))
        self.primLegendLabelNormal.setText(QCoreApplication.translate("PrimLegend", u"Normal", None))
        self.primLegendLabelDimmed.setText(QCoreApplication.translate("PrimLegend", u"Dimmed colors denote inactive prims", None))
        self.primLegendLabelFontsAbstract.setText(QCoreApplication.translate("PrimLegend", u"Normal font indicates abstract prims(class and children)", None))
        self.primLegendLabelFontsUndefined.setText(QCoreApplication.translate("PrimLegend", u"Italic font indicates undefined prims(declared with over)", None))
        self.primLegendLabelFontsDefined.setText(QCoreApplication.translate("PrimLegend", u"Bold font indicates defined prims(declared with def)", None))
    # retranslateUi




#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from __future__ import print_function

from .qt import QtGui, QtWidgets
from .usdviewContextMenuItem import UsdviewContextMenuItem
import os
import sys

#
# Edit the following to alter the per-prim context menu.
#
# Every entry should be an object derived from PrimContextMenuItem,
# defined below.
#
def _GetContextMenuItems(appController, item):
    return [JumpToEnclosingModelItem(appController, item),
            SelectBoundPreviewMaterialMenuItem(appController, item),
            SelectBoundFullMaterialMenuItem(appController, item),
            SeparatorMenuItem(appController, item),
            ToggleVisibilityMenuItem(appController, item),
            VisOnlyMenuItem(appController, item),
            RemoveVisMenuItem(appController, item),
            SeparatorMenuItem(appController, item),
            LoadOrUnloadMenuItem(appController, item),
            ActiveMenuItem(appController, item),
            SeparatorMenuItem(appController, item),
            CopyPrimPathMenuItem(appController, item),
            CopyModelPathMenuItem(appController, item),
            SeparatorMenuItem(appController, item),
            IsolateAssetMenuItem(appController, item),
            SetAsActiveCamera(appController, item),
            SetAsActiveRenderSettings(appController, item),
            SetAsActiveRenderPassPrim(appController, item)]

#
# The base class for per-prim context menu items.
#
class PrimContextMenuItem(UsdviewContextMenuItem):

    def __init__(self, appController, item):
        self._selectionDataModel = appController._dataModel.selection
        self._currentFrame = appController._dataModel.currentFrame
        self._appController = appController
        self._item = item

    def IsEnabled(self):
        return True

    def IsSeparator(self):
        return False

    def GetText(self):
        return ""

    def RunCommand(self):
        return True

#
# Puts a separator in the context menu
#
class SeparatorMenuItem(PrimContextMenuItem):

    def IsSeparator(self):
        return True

#
# Replace each selected prim with its enclosing model prim, if it has one,
# in the selection set
#
class JumpToEnclosingModelItem(PrimContextMenuItem):

    def IsEnabled(self):
        from .common import GetEnclosingModelPrim

        for p in self._selectionDataModel.getPrims():
            if GetEnclosingModelPrim(p) is not None:
                return True
        return False

    def GetText(self):
        return "Jump to Enclosing Model"

    def RunCommand(self):
        self._appController.selectEnclosingModel()

#
# Replace each selected prim with the "preview" Material it is bound to.
#
class SelectBoundPreviewMaterialMenuItem(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)
        from pxr import UsdShade 

        self._boundPreviewMaterial = None
        self._bindingRel = None
        for p in self._selectionDataModel.getPrims():
            (self._boundPreviewMaterial, self._bindingRel) = \
                UsdShade.MaterialBindingAPI(p).ComputeBoundMaterial(
                    UsdShade.Tokens.preview)
            if self._boundPreviewMaterial:
                break

    def IsEnabled(self):
        return bool(self._boundPreviewMaterial)

    def GetText(self):
        if self._boundPreviewMaterial:
            isPreviewBindingRel = 'preview' in self._bindingRel.SplitName()
            return "Select Bound Preview Material (%s%s)" % (
            self._boundPreviewMaterial.GetPrim().GetName(),
            "" if isPreviewBindingRel else " from generic binding")
        else:
            return "Select Bound Preview Material (None)"

    def RunCommand(self):
        self._appController.selectBoundPreviewMaterial()

#
# Replace each selected prim with the "preview" Material it is bound to.
#
class SelectBoundFullMaterialMenuItem(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)
        from pxr import UsdShade 

        self._boundFullMaterial = None
        self._bindingRel = None
        for p in self._selectionDataModel.getPrims():
            (self._boundFullMaterial, self._bindingRel) = \
                UsdShade.MaterialBindingAPI(p).ComputeBoundMaterial(
                    UsdShade.Tokens.full)
            if self._boundFullMaterial:
                break

    def IsEnabled(self):
        return bool(self._boundFullMaterial)

    def GetText(self):
        if self._boundFullMaterial:
            isFullBindingRel = 'full' in self._bindingRel.SplitName()
            return "Select Bound Full Material (%s%s)" % (
            self._boundFullMaterial.GetPrim().GetName(),
            "" if isFullBindingRel else " from generic binding")
        else:
            return "Select Bound Full Material (None)"

    def RunCommand(self):
        self._appController.selectBoundFullMaterial()


#
# Allows you to activate/deactivate a prim in the graph
#
class ActiveMenuItem(PrimContextMenuItem):

    def GetText(self):
        if self._selectionDataModel.getFocusPrim().IsActive():
            return "Deactivate"
        else:
            return "Activate"

    def RunCommand(self):
        active = self._selectionDataModel.getFocusPrim().IsActive()
        if active:
            self._appController.deactivateSelectedPrims()
        else:
            self._appController.activateSelectedPrims()

#
# Allows you to vis or invis a prim in the graph, based on its current
# resolved visibility
#
class ToggleVisibilityMenuItem(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)
        from pxr import UsdGeom
        self._imageable = False
        self._isVisible = False
        for prim in self._selectionDataModel.getPrims():
            imgbl = UsdGeom.Imageable(prim)
            if imgbl:
                self._imageable = True
                self._isVisible = (imgbl.ComputeVisibility(self._currentFrame)
                                   == UsdGeom.Tokens.inherited)
                break

    def IsEnabled(self):
        return self._imageable

    def GetText(self):
        return "Make Invisible" if self._isVisible else "Make Visible"

    def RunCommand(self):
        if self._isVisible:
            self._appController.invisSelectedPrims()
        else:
            self._appController.visSelectedPrims()


#
# Allows you to vis-only a prim in the graph
#
class VisOnlyMenuItem(PrimContextMenuItem):

    def IsEnabled(self):
        from pxr import UsdGeom
        for prim in self._selectionDataModel.getPrims():
            if prim.IsA(UsdGeom.Imageable):
                return True
        return False

    def GetText(self):
        return "Vis Only"

    def RunCommand(self):
        self._appController.visOnlySelectedPrims()

#
# Remove any vis/invis authored on selected prims
#
class RemoveVisMenuItem(PrimContextMenuItem):

    def IsEnabled(self):
        from .common import HasSessionVis
        for prim in self._selectionDataModel.getPrims():
            if HasSessionVis(prim):
                return True

        return False

    def GetText(self):
        return "Remove Session Visibility"

    def RunCommand(self):
        self._appController.removeVisSelectedPrims()


#
# Toggle load-state on the selected prims, if loadable
#
class LoadOrUnloadMenuItem(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)
        from .common import GetPrimsLoadability
        # Use the descendent-pruned selection set to avoid redundant
        # traversal of the stage to answer isLoaded...
        self._loadable, self._loaded = GetPrimsLoadability(
            self._selectionDataModel.getLCDPrims())

    def IsEnabled(self):
        return self._loadable

    def GetText(self):
        return "Unload" if self._loaded else "Load"

    def RunCommand(self):
        if self._loaded:
            self._appController.unloadSelectedPrims()
        else:
            self._appController.loadSelectedPrims()



#
# Copies the paths of the currently selected prims to the clipboard
#
class CopyPrimPathMenuItem(PrimContextMenuItem):

    def GetText(self):
        if len(self._selectionDataModel.getPrims()) > 1:
            return "Copy Prim Paths"
        return "Copy Prim Path"

    def RunCommand(self):
        pathlist = [str(p.GetPath())
            for p in self._selectionDataModel.getPrims()]
        pathStrings = '\n'.join(pathlist)

        cb = QtWidgets.QApplication.clipboard()
        cb.setText(pathStrings, QtGui.QClipboard.Mode.Selection )
        cb.setText(pathStrings, QtGui.QClipboard.Mode.Clipboard )

#
#  Copies the path of the first-selected prim's enclosing model
#  to the clipboard, if the prim is inside a model
#
class CopyModelPathMenuItem(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)
        from .common import GetEnclosingModelPrim

        if len(self._selectionDataModel.getPrims()) == 1:
            self._modelPrim = GetEnclosingModelPrim(
                self._selectionDataModel.getFocusPrim())
        else:
            self._modelPrim = None

    def IsEnabled(self):
        return self._modelPrim

    def GetText(self):
        name = ( "(%s)" % self._modelPrim.GetName() ) if self._modelPrim else ""
        return "Copy Enclosing Model %s Path" % name

    def RunCommand(self):
        modelPath = str(self._modelPrim.GetPath())
        cb = QtWidgets.QApplication.clipboard()
        cb.setText(modelPath, QtGui.QClipboard.Mode.Selection )
        cb.setText(modelPath, QtGui.QClipboard.Mode.Clipboard )



#
# Copies the current prim and subtree to a file of the user's choosing
# XXX This is not used, and does not work. Leaving code in for now for
# future reference/inspiration
#
class IsolateCopyPrimMenuItem(PrimContextMenuItem):

    def GetText(self):
        return "Isolate Copy of Prim..."

    def RunCommand(self):
        focusPrim = self._selectionDataModel.getFocusPrim()

        inFile = focusPrim.GetScene().GetUsdFile()

        guessOutFile = os.getcwd() + "/" + focusPrim.GetName() + "_copy.usd"
        (outFile, _) = QtWidgets.QFileDialog.getSaveFileName(None,
            "Specify the Usd file to create", guessOutFile, 'Usd files (*.usd)')
        if (outFile.rsplit('.')[-1] != 'usd'):
            outFile += '.usd'

        if inFile == outFile:
            sys.stderr.write( "Cannot isolate a copy to the source usd!\n" )
            return

        sys.stdout.write( "Writing copy to new file '%s' ... " % outFile )
        sys.stdout.flush()

        os.system( 'usdcopy -inUsd ' + inFile +
                ' -outUsd ' + outFile + ' ' +
                ' -sourcePath ' + focusPrim.GetPath() + '; ' +
                'usdview ' + outFile + ' &')

        sys.stdout.write( "Done!\n" )

    def IsEnabled(self):
        numSelectedPrims = len(self._selectionDataModel.getPrims())
        focusPrimActive = self._selectionDataModel.getFocusPrim().GetActive()

        return numSelectedPrims == 1 and focusPrimActive


#
# Launches usdview on the asset instantiated at the selected prim, as
# defined by USD assetInfo present on the prim
#
class IsolateAssetMenuItem(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)

        self._assetName = None
        if len(self._selectionDataModel.getPrims()) == 1:
            from pxr import Usd
            model = Usd.ModelAPI(self._selectionDataModel.getFocusPrim())
            name = model.GetAssetName()
            identifier = model.GetAssetIdentifier()
            if name and identifier:
                # Ar API is still settling out...
                # from pxr import Ar
                # identifier = Ar.GetResolver().Resolve("", identifier.path)
                from pxr import Sdf
                layer = Sdf.Layer.Find(identifier.path)
                if layer:
                    self._assetName = name
                    self._filePath = layer.realPath

    def IsEnabled(self):
        return self._assetName

    def GetText(self):
        name = ( " '%s'" % self._assetName ) if self._assetName else ""
        return "usdview asset%s" % name

    def RunCommand(self):
        print("Spawning usdview %s" % self._filePath)
        os.system("usdview %s &" % self._filePath)

#
# If the selected prim is a camera and not the currently active camera, display
# an enabled menu item to set it as the active camera.
#
class SetAsActiveCamera(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)

        self._nonActiveCameraPrim = None
        if len(self._selectionDataModel.getPrims()) == 1:
            prim = self._selectionDataModel.getPrims()[0]
            from pxr import UsdGeom
            if UsdGeom.Camera(prim):
                if prim != appController.getActiveCamera():
                    self._nonActiveCameraPrim = prim

    def IsEnabled(self):
        return self._nonActiveCameraPrim

    def GetText(self):
        return "Set As Active Camera"

    def RunCommand(self):
        self._appController._cameraSelectionChanged(self._nonActiveCameraPrim)

#
# If the selected prim is a render settings prim and not the currently active
# render settings prim, display an enabled menu item to set it as the active
# render settings prim.
#
class SetAsActiveRenderSettings(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)

        self._nonActiveRenderSettingsPrim = None
        if len(self._selectionDataModel.getPrims()) == 1:
            prim = self._selectionDataModel.getPrims()[0]
            from pxr import UsdRender
            if UsdRender.Settings(prim):
                if prim != appController.getActiveRenderSettingsPrim():
                    self._nonActiveRenderSettingsPrim = prim

    def IsEnabled(self):
        return self._nonActiveRenderSettingsPrim

    def GetText(self):
        return "Set As Active Render Settings"

    def RunCommand(self):
        self._appController.setActiveRenderSettingsPrim(
            self._nonActiveRenderSettingsPrim)

#
# If the selected prim is a render pass prim and not the currently active
# render pass prim, display an enabled menu item to set it as the active
# render pass prim.
#
class SetAsActiveRenderPassPrim(PrimContextMenuItem):

    def __init__(self, appController, item):
        PrimContextMenuItem.__init__(self, appController, item)

        self._nonActiveRenderPassPrim = None
        if len(self._selectionDataModel.getPrims()) == 1:
            prim = self._selectionDataModel.getPrims()[0]
            from pxr import UsdRender
            if UsdRender.Pass(prim):
                if prim != appController.getActiveRenderPassPrim():
                    self._nonActiveRenderPassPrim = prim

    def IsEnabled(self):
        return self._nonActiveRenderPassPrim

    def GetText(self):
        return "Set As Active Render Pass Prim"

    def RunCommand(self):
        self._appController.setActiveRenderPassPrim(
            self._nonActiveRenderPassPrim)


#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

# pylint: disable=map-builtin-not-iterating

from .qt import QtCore, QtGui, QtWidgets


def _GetLengthOfRange(start, stop, step):
    if step == 0:
        return 0

    k = 1 if step > 0 else -1
    num = 1 + (stop - k - start) // step

    # note, sign(stop - start) != sign(step) when we have an empty range.  In those
    # cases, "num" ends up non-positive, so we can just take the max(num, 0)
    return max(0, num)


class _ArrayAttributeModel(QtCore.QAbstractListModel):
    '''This is a data model that represents a slice into some array data.
    '''

    RawDataRole = QtCore.Qt.ItemDataRole.UserRole + 0

    def __init__(self):
        super(_ArrayAttributeModel, self).__init__()
        self._arrayData = None
        self._scalarTypeName = ""
        self._slice = slice(None)
        self._fetchMoreTimer = QtCore.QTimer()
        self._fetchMoreTimer.setInterval(1000)
        self._fetchMoreTimer.timeout.connect(self.TryToFetchMore)
        self._fetchMoreTimer.start()

        self._rowCount = 0
        self._publishedRows = 0

    def SetArrayDataAndTypeName(self, arrayData, scalarTypeName):
        self._arrayData = arrayData
        self._scalarTypeName = scalarTypeName
        self._Reset()

    def SetSlice(self, slice_):
        if self._slice == slice_:
            return
        self._slice = slice_
        self._Reset()

    def _Reset(self):
        self._publishedRows = 0
        self._rowCount = 0
        if self._arrayData is not None:
            self._rowCount = _GetLengthOfRange(*self._slice.indices(len(self._arrayData)))
        self.modelReset.emit()

    def GetArrayData(self):
        return self._arrayData

    def GetScalarTypeName(self):
        return self._scalarTypeName

    def index(self, row, col, parent=QtCore.QModelIndex()):
        return self.createIndex(row, col)

    def parent(self, index):
        return QtCore.QModelIndex()

    def rowCount(self, parent=QtCore.QModelIndex()):
        if not parent.isValid():
            return self._publishedRows
        return 0

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 1

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        start, _, step = self._slice.indices(len(self._arrayData))
        idx = start + index.row() * step
        dataVal = self._arrayData[idx]

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            from .scalarTypes import ToString
            return str(idx) + ": " + ToString(
                dataVal, self._scalarTypeName)

        elif role == _ArrayAttributeModel.RawDataRole:
            return dataVal

        return None

    def fetchMore(self, index):
        left = self._rowCount - self._publishedRows
        toFetch = min(5000 if self._publishedRows == 0 else 100, left)
        self.beginInsertRows(
            index, self._publishedRows, self._publishedRows + toFetch - 1)
        self._publishedRows += toFetch
        self.endInsertRows()

    def canFetchMore(self, index):
        return self._publishedRows < self._rowCount

    def TryToFetchMore(self):
        rootIndex = QtCore.QModelIndex()
        if self.canFetchMore(rootIndex):
            self.fetchMore(rootIndex)


class ArrayAttributeView(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super(ArrayAttributeView, self).__init__(parent)

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        sliceLayout = QtWidgets.QHBoxLayout()
        layout.addLayout(sliceLayout)

        sliceLayout.addWidget(QtWidgets.QLabel("Slice:"))
        self._lineEdit = _SliceLineEdit()
        self._lineEdit.setPlaceholderText('e.g.: "0:5000", "::-1"')
        sliceLayout.addWidget(self._lineEdit)

        self._arrayAttrModel = _ArrayAttributeModel()
        self._listView = QtWidgets.QListView()
        self._listView.setUniformItemSizes(True)
        self._listView.setViewMode(QtWidgets.QListView.ViewMode.ListMode)
        self._listView.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self._listView.setModel(self._arrayAttrModel)
        layout.addWidget(self._listView)

        self._lineEdit.SliceChanged.connect(self._arrayAttrModel.SetSlice)

        self._SetupContextMenu()

    def SetAttribute(self, attr, frame):
        from .scalarTypes import GetScalarTypeFromAttr

        arrayData = attr.Get(frame)
        scalarTypeName, _ = GetScalarTypeFromAttr(attr)
        self._arrayAttrModel.SetArrayDataAndTypeName(arrayData, scalarTypeName)

    def CanView(self, attr):
        return attr.GetTypeName().isArray

    def Clear(self):
        self._arrayAttrModel.SetArrayDataAndTypeName(None, "")
        self._lineEdit.setText("")

    def keyPressEvent(self, e):
        # XXX note, this is extremely finicky.  it does not really
        # have keyboard focus.
        if e.matches(QtGui.QKeySequence.StandardKey.Copy):
            self.Copy()
        else:
            return super(ArrayAttributeView, self).keyPressEvent(e)

    # context menu stuff
    def _SetupContextMenu(self):
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._ShowContextMenu)

    def _ShowContextMenu(self, point):
        menu = QtWidgets.QMenu(self)
        if self._listView.selectedIndexes():
            menu.addAction("Copy Selected", self.CopySelected)
        else:
            menu.addAction("Copy All", self.CopyAll)
        menu.addAction("Select All", self.SelectAll)
        menu.exec_(QtGui.QCursor.pos())

    def CopyAll(self):
        self._CopyValsToClipboard(self._arrayAttrModel.GetArrayData())

    def CopySelected(self):
        selectedIndexes = self._listView.selectedIndexes()
        self._CopyValsToClipboard([
            self._listView.model().data(idx, _ArrayAttributeModel.RawDataRole)
            for idx in selectedIndexes])

    def _CopyValsToClipboard(self, vals):
        from .scalarTypes import ToClipboard
        scalarTypeName = self._arrayAttrModel.GetScalarTypeName()
        copyText = "[ %s ]" % (", ".join(
            ToClipboard(val, scalarTypeName)
            for val in vals))
        QtWidgets.QApplication.clipboard().setText(copyText)

    def SelectAll(self):
        self._listView.selectAll()


def _IntOrNone(s):
    if not s:
        return None
    return int(s)


def _GetSliceFromString(s):
    s = slice(*map(_IntOrNone, s.split(':')))
    if s.step == 0:
        raise ValueError("Slice cannot be 0")
    return s


class _SliceLineEdit(QtWidgets.QLineEdit):
    """LineEdit for inputing strings that represent slices"""

    SliceChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super(_SliceLineEdit, self).__init__(parent)
        self.setValidator(_SliceLineEdit.Validator())
        self.editingFinished.connect(self._OnEditingFinished)

    class Validator(QtGui.QValidator):
        def validate(self, s, pos):
            s = s.strip()
            try:
                _GetSliceFromString(s)
                return (QtGui.QValidator.State.Acceptable, s, pos)
            except:
                return (QtGui.QValidator.State.Intermediate, s, pos)

    def setText(self, t):
        super(_SliceLineEdit, self).setText(t)
        self._OnEditingFinished()

    def _OnEditingFinished(self):
        self.SliceChanged.emit(_GetSliceFromString(self.text()))



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

# pylint: disable=round-builtin

from __future__ import division
from __future__ import print_function

from .qt import QtCore, QtGui, QtWidgets
import os, time, sys, platform, math
from pxr import Ar, Tf, Trace, Sdf, Kind, Usd, UsdGeom, UsdShade
from .customAttributes import CustomAttribute
from pxr.UsdUtils.constantsGroup import ConstantsGroup

DEBUG_CLIPPING = "USDVIEWQ_DEBUG_CLIPPING"

class ClearColors(ConstantsGroup):
    """Names of available background colors."""
    BLACK = "Black"
    DARK_GREY = "Grey (Dark)"
    LIGHT_GREY = "Grey (Light)"
    WHITE = "White"

class DefaultFontFamily(ConstantsGroup):
    """Names of the default font family and monospace font family to be used
    with usdview"""
    FONT_FAMILY = "Roboto"
    MONOSPACE_FONT_FAMILY = "Roboto Mono"

class HighlightColors(ConstantsGroup):
    """Names of available highlight colors for selected objects."""
    WHITE = "White"
    YELLOW = "Yellow"
    CYAN = "Cyan"

class UIBaseColors(ConstantsGroup):
    RED = QtGui.QBrush(QtGui.QColor(230, 132, 131))
    LIGHT_SKY_BLUE = QtGui.QBrush(QtGui.QColor(135, 206, 250))
    DARK_YELLOW = QtGui.QBrush(QtGui.QColor(222, 158, 46))

class UIPrimTypeColors(ConstantsGroup):
    HAS_ARCS = UIBaseColors.DARK_YELLOW
    NORMAL = QtGui.QBrush(QtGui.QColor(227, 227, 227))
    INSTANCE = UIBaseColors.LIGHT_SKY_BLUE
    PROTOTYPE = QtGui.QBrush(QtGui.QColor(118, 136, 217))

class UIPrimTreeColors(ConstantsGroup):
    SELECTED = QtGui.QBrush(QtGui.QColor(189, 155, 84))
    SELECTED_HOVER = QtGui.QBrush(QtGui.QColor(227, 186, 101))
    ANCESTOR_OF_SELECTED = QtGui.QBrush(QtGui.QColor(189, 155, 84, 50))
    ANCESTOR_OF_SELECTED_HOVER = QtGui.QBrush(QtGui.QColor(189, 155, 84, 100))
    UNSELECTED_HOVER = QtGui.QBrush(QtGui.QColor(70, 70, 70))

class UIPropertyValueSourceColors(ConstantsGroup):
    FALLBACK = UIBaseColors.DARK_YELLOW
    TIME_SAMPLE = QtGui.QBrush(QtGui.QColor(177, 207, 153))
    DEFAULT = UIBaseColors.LIGHT_SKY_BLUE
    NONE = QtGui.QBrush(QtGui.QColor(140, 140, 140))
    VALUE_CLIPS = QtGui.QBrush(QtGui.QColor(230, 150, 230))
    SPLINE = QtGui.QBrush(QtGui.QColor(80, 200, 180))

class UIFonts(ConstantsGroup):
    # Font constants.  We use font in the prim browser to distinguish
    # "resolved" prim specifier
    # XXX - the use of weight here may need to be revised depending on font family
    BASE_POINT_SIZE = 10
    
    ITALIC = QtGui.QFont()
    ITALIC.setWeight(QtGui.QFont.Weight.Light)
    ITALIC.setItalic(True)

    NORMAL = QtGui.QFont()
    NORMAL.setWeight(QtGui.QFont.Weight.Normal)

    BOLD = QtGui.QFont()
    BOLD.setWeight(QtGui.QFont.Weight.Bold)

    BOLD_ITALIC = QtGui.QFont()
    BOLD_ITALIC.setWeight(QtGui.QFont.Weight.Bold)
    BOLD_ITALIC.setItalic(True)

    OVER_PRIM = ITALIC
    DEFINED_PRIM = BOLD
    ABSTRACT_PRIM = NORMAL

    INHERITED = QtGui.QFont()
    INHERITED.setPointSizeF(BASE_POINT_SIZE * 0.8)
    INHERITED.setWeight(QtGui.QFont.Weight.Normal)
    INHERITED.setItalic(True)

class KeyboardShortcuts(ConstantsGroup):
    FramingKey = QtCore.Qt.Key.Key_F

class PropertyViewIndex(ConstantsGroup):
    TYPE, NAME, VALUE = range(3)

def GetIconPath(name):
    iconDir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'icons')
    return os.path.join(iconDir, name)

# We use deferred loading because icons can't be constructed before
# application initialization time.
_icons = {}
def _DeferredIconLoad(name):
    fullPath = GetIconPath(name)
    try:
        icon = _icons[fullPath]
    except KeyError:
        icon = QtGui.QIcon(fullPath)
        _icons[fullPath] = icon
    return icon

class PropertyViewIcons(ConstantsGroup):
    ATTRIBUTE                  = lambda: _DeferredIconLoad('usd-attr-plain-icon.svg')
    ATTRIBUTE_WITH_CONNECTIONS = lambda: _DeferredIconLoad('usd-attr-with-conn-icon.svg')
    RELATIONSHIP               = lambda: _DeferredIconLoad('usd-rel-plain-icon.svg')
    RELATIONSHIP_WITH_TARGETS  = lambda: _DeferredIconLoad('usd-rel-with-target-icon.svg')
    TARGET                     = lambda: _DeferredIconLoad('usd-target-icon.svg')
    CONNECTION                 = lambda: _DeferredIconLoad('usd-conn-icon.svg')
    COMPOSED                   = lambda: _DeferredIconLoad('usd-cmp-icon.svg')

class PropertyViewDataRoles(ConstantsGroup):
    ATTRIBUTE = "Attr"
    RELATIONSHIP = "Rel"
    ATTRIBUTE_WITH_CONNNECTIONS = "Attr_"
    RELATIONSHIP_WITH_TARGETS = "Rel_"
    TARGET = "Tgt"
    CONNECTION = "Conn"
    COMPOSED = "Cmp"
    NORMALIZED_NAME = QtCore.Qt.ItemDataRole.UserRole + 1

class RenderModes(ConstantsGroup):
    # Render modes
    WIREFRAME = "Wireframe"
    WIREFRAME_ON_SURFACE = "WireframeOnSurface"
    SMOOTH_SHADED = "Smooth Shaded"
    FLAT_SHADED = "Flat Shaded"
    POINTS = "Points"
    GEOM_ONLY = "Geom Only"
    GEOM_FLAT = "Geom Flat"
    GEOM_SMOOTH = "Geom Smooth"
    HIDDEN_SURFACE_WIREFRAME = "Hidden Surface Wireframe"

class ShadedRenderModes(ConstantsGroup):
    # Render modes which use shading
    SMOOTH_SHADED = RenderModes.SMOOTH_SHADED
    FLAT_SHADED = RenderModes.FLAT_SHADED
    WIREFRAME_ON_SURFACE = RenderModes.WIREFRAME_ON_SURFACE
    GEOM_FLAT = RenderModes.GEOM_FLAT
    GEOM_SMOOTH = RenderModes.GEOM_SMOOTH

class ColorCorrectionModes(ConstantsGroup):
    # Color correction used when render is presented to screen
    # These strings should match HdxColorCorrectionTokens
    DISABLED = "disabled"
    SRGB = "sRGB"
    OPENCOLORIO = "openColorIO"

class PickModes(ConstantsGroup):
    # Pick modes
    PRIMS = "Select Prims"
    MODELS = "Select Models"
    INSTANCES = "Select Instances"
    PROTOTYPES = "Select Prototypes"

class SelectionHighlightModes(ConstantsGroup):
    # Selection highlight modes
    NEVER = "Never"
    ONLY_WHEN_PAUSED = "Only when paused"
    ALWAYS = "Always"

class CameraMaskModes(ConstantsGroup):
    NONE = "none"
    PARTIAL = "partial"
    FULL = "full"

class IncludedPurposes(ConstantsGroup):
    DEFAULT = UsdGeom.Tokens.default_
    PROXY = UsdGeom.Tokens.proxy
    GUIDE = UsdGeom.Tokens.guide
    RENDER = UsdGeom.Tokens.render

def _PropTreeWidgetGetRole(tw):
    return tw.data(PropertyViewIndex.TYPE, QtCore.Qt.ItemDataRole.WhatsThisRole)

def PropTreeWidgetTypeIsRel(tw):
    role = _PropTreeWidgetGetRole(tw)
    return role in (PropertyViewDataRoles.RELATIONSHIP,
                    PropertyViewDataRoles.RELATIONSHIP_WITH_TARGETS)

def _UpdateLabelText(text, substring, mode):
    return text.replace(substring, '<'+mode+'>'+substring+'</'+mode+'>')

def ItalicizeLabelText(text, substring):
    return _UpdateLabelText(text, substring, 'i')

def BoldenLabelText(text, substring):
    return _UpdateLabelText(text, substring, 'b')

def ColorizeLabelText(text, substring, r, g, b):
    return _UpdateLabelText(text, substring,
                            "span style=\"color:rgb(%d, %d, %d);\"" % (r, g, b))

def PrintWarning(title, description):
    msg = sys.stderr
    print("------------------------------------------------------------", file=msg)
    print("WARNING: %s" % title, file=msg)
    print(description, file=msg)
    print("------------------------------------------------------------", file=msg)

def GetValueAndDisplayString(prop, time):
    """If `prop` is a timeSampled Sdf.AttributeSpec, compute a string specifying
    how many timeSamples it possesses.  Otherwise, compute the single default
    value, or targets for a relationship, or value at 'time' for a
    Usd.Attribute.  Return a tuple of a parameterless function that returns the
    resolved value at 'time', and the computed brief string for display.  We
    return a value-producing function rather than the value itself because for
    an Sdf.AttributeSpec with multiple timeSamples, the resolved value is
    *all* of the timeSamples, which can be expensive to compute, and is
    rarely needed.
    """
    def _ValAndStr(val): 
        return (lambda: val, GetShortStringForValue(prop, val))

    if isinstance(prop, Usd.Relationship):
        return _ValAndStr(prop.GetTargets())
    elif isinstance(prop, (Usd.Attribute, CustomAttribute)):
        return _ValAndStr(prop.Get(time))
    elif isinstance(prop, Sdf.AttributeSpec):
        if time == Usd.TimeCode.Default():
            return _ValAndStr(prop.default)
        else:
            numTimeSamples = prop.layer.GetNumTimeSamplesForPath(prop.path)
            if numTimeSamples == 0:
                return _ValAndStr(prop.default)
            else:
                def _GetAllTimeSamples(attrSpec):
                    l = attrSpec.layer
                    p = attrSpec.path
                    ordinates = l.ListTimeSamplesForPath(p)
                    return [(o, l.QueryTimeSample(p, o)) for o in ordinates]

                if numTimeSamples == 1:
                    valStr = "1 time sample"
                else:
                    valStr = str(numTimeSamples) + " time samples"
                    
                return (lambda prop=prop: _GetAllTimeSamples(prop), valStr)

    elif isinstance(prop, Sdf.RelationshipSpec):
        return _ValAndStr(prop.targetPathList)
    
    return (lambda: None, "unrecognized property type")


def GetShortStringForValue(prop, val):
    if isinstance(prop, Usd.Relationship):
        val = ", ".join(str(p) for p in val)
    elif isinstance(prop, Sdf.RelationshipSpec):
        return str(prop.targetPathList)

    # If there is no value opinion, we do not want to display anything,
    # since python 'None' has a different meaning than usda-authored None,
    # which is how we encode attribute value blocks (which evaluate to 
    # Sdf.ValueBlock)
    if val is None:
        return ''
    
    valType = Sdf.GetValueTypeNameForValue(val)
    result = ''
    if valType.isArray and not isinstance(val, Sdf.ValueBlock):
        def arrayToStr(a):
            from itertools import chain
            elems = a if len(a) <= 6 else chain(a[:3], ['...'], a[-3:])
            return '[' + ', '.join(map(str, elems)) + ']'
        if val is not None and len(val):
            result = "%s[%d]: %s" % (
                valType.scalarType, len(val), arrayToStr(val))
        else:
            result = "%s[]" % valType.scalarType
    else:
        result = str(val)

    return result[:500]

# Return a string that reports size in metric units (units of 1000, not 1024).
def ReportMetricSize(sizeInBytes):
    if sizeInBytes == 0:
       return "0 B"
    sizeSuffixes = ("B", "KB", "MB", "GB", "TB", "PB", "EB")
    i = int(math.floor(math.log(sizeInBytes, 1000)))
    if i >= len(sizeSuffixes):
        i = len(sizeSuffixes) - 1
    p = math.pow(1000, i)
    s = round(sizeInBytes / p, 2)
    return "%s %s" % (s, sizeSuffixes[i])

# Return attribute status at a certian frame (is it using the default, or the
# fallback? Is it authored at this frame? etc.
def _GetAttributeStatus(attribute, frame):
    return attribute.GetResolveInfo(frame).GetSource()

# Return a Font corresponding to certain attribute properties.
# Currently this only applies italicization on interpolated time samples.
def GetPropertyTextFont(prop, frame):
    if not isinstance(prop, Usd.Attribute):
        # Early-out for non-attribute properties.
        return None

    frameVal = frame.GetValue()
    bracketing = prop.GetBracketingTimeSamples(frameVal)

    # Note that some attributes return an empty tuple, some None, from
    # GetBracketingTimeSamples(), but all will be fed into this function.
    if bracketing and (len(bracketing) == 2) and (bracketing[0] != frameVal):
        return UIFonts.ITALIC

    if prop.HasSpline():
        spline = prop.GetSpline()
        if frameVal not in spline.GetKnots().keys():
            return UIFonts.ITALIC

    return None

# Helper function that takes attribute status and returns the display color
def GetPropertyColor(prop, frame, hasValue=None, hasAuthoredValue=None,
                      valueIsDefault=None):
    if not isinstance(prop, Usd.Attribute):
        # Early-out for non-attribute properties.
        return UIBaseColors.RED.color()

    statusToColor = {Usd.ResolveInfoSourceFallback   : UIPropertyValueSourceColors.FALLBACK,
                     Usd.ResolveInfoSourceDefault    : UIPropertyValueSourceColors.DEFAULT,
                     Usd.ResolveInfoSourceValueClips : UIPropertyValueSourceColors.VALUE_CLIPS,
                     Usd.ResolveInfoSourceTimeSamples: UIPropertyValueSourceColors.TIME_SAMPLE,
                     Usd.ResolveInfoSourceNone       : UIPropertyValueSourceColors.NONE,
                     Usd.ResolveInfoSourceSpline     : UIPropertyValueSourceColors.SPLINE}

    valueSource = _GetAttributeStatus(prop, frame)

    return statusToColor[valueSource].color()

# Gathers information about a layer used as a subLayer, including its
# position in the layerStack hierarchy.
class LayerInfo(object):
    def __init__(self, identifier, realPath, offset, stage, 
                 timeCodesPerSecond=None, isMuted=False, depth=0):
        self._identifier = identifier
        self._realPath = realPath
        self._offset = offset
        self._stage = stage
        self._timeCodesPerSecond = timeCodesPerSecond
        self._isMuted = isMuted
        self._depth = depth

    @classmethod
    def FromLayer(cls, layer, stage, offset, depth=0):
        return cls(layer.identifier, layer.realPath, offset, stage,
                   timeCodesPerSecond=layer.timeCodesPerSecond,
                   depth=depth)

    @classmethod
    def FromMutedLayerIdentifier(cls, identifier, parentLayer, stage, depth=0):
        realPath = ''
        try:
            resolver = Ar.GetResolver()
            realPath = resolver.Resolve(identifier).GetPathString()
        except Exception as e:
            PrintWarning('Failed to resolve identifier {} '
                         .format(identifier), e)
            realPath = 'unknown'

        return cls(identifier, realPath, Sdf.LayerOffset(), stage, 
                   isMuted=True, depth=depth)

    def GetIdentifier(self):
        return self._identifier

    def GetRealPath(self):
        return self._realPath
            
    def IsMuted(self):
        return self._isMuted

    def GetOffset(self):
        return self._offset
    
    def GetOffsetString(self):
        if self._offset == None:
            return '-'
        if self._offset.IsIdentity():
            return ""
        else:
            return "{} , {}".format(self._offset.offset, self._offset.scale)

    def GetOffsetTooltipString(self):
        if self._offset == None:
            return '-'
        if self._offset.IsIdentity():
            return ""
        toolTips = ["<b>Layer Offset</b> (from stage root)",
                    "<b>offset:</b> {}".format(self._offset.offset),
                    "<b>scale:</b> {}".format(self._offset.scale)]
        # Display info about automatic time scaling if the layer's tcps is known
        # and doesn't match the stage's tcps.
        if self._timeCodesPerSecond:
            stageTcps = self._stage.GetTimeCodesPerSecond()
            if self._timeCodesPerSecond != stageTcps:
                toolTips.append("Includes timeCodesPerSecond auto-scaling: "
                                "{} (stage) / {} (layer)".format(
                    stageTcps, self._timeCodesPerSecond))
        return "<br>".join(toolTips)

    def GetToolTipString(self):
        return "<b>identifier:</b> @%s@ <br> <b>resolved path:</b> %s" % \
            (self.GetIdentifier(), self.GetRealPath())

    def GetHierarchicalDisplayString(self):
        return ('    '*self._depth + 
            Sdf.Layer.GetDisplayNameFromIdentifier(self._identifier))

def _AddLayerTree(stage, layerTree, depth=0):

    layers = [LayerInfo.FromLayer(
        layerTree.layer, stage, layerTree.offset, depth)]

    # The layer tree from the layer stack has all of the fully composed layer
    # offsets, but will not have any of the muted layers. The sublayer paths of
    # layer will still contain any muted layers but will not have the composed
    # layer offsets that the layer tree provides. So in order to show the muted
    # layers in the correct sublayer position, we go through the sublayer paths
    # parsing either the muted layer or a layer stack tree subtree.
    # 
    # The layer tree will also not have any entries for layers that failed to
    # load. We need to handle this case as well.
    childTrees = layerTree.childTrees
    subLayerPaths = layerTree.layer.subLayerPaths
    childTreeIter = iter(layerTree.childTrees)
    numMissingLayers = 0
    for subLayerPath in subLayerPaths:
        anchoredSubLayerPath = Sdf.ComputeAssetPathRelativeToLayer(
            layerTree.layer, subLayerPath)
        if stage.IsLayerMuted(anchoredSubLayerPath):
            # The sublayer path is muted so add muted layer by path. We don't 
            # recurse on sublayers for muted layers.
            layers.append(LayerInfo.FromMutedLayerIdentifier(
                anchoredSubLayerPath, layerTree.layer, stage, depth=depth+1))
            numMissingLayers = numMissingLayers + 1
        elif not Sdf.Layer.Find(anchoredSubLayerPath):
            # Otherwise, the sublayer failed to load for some other reason.
            # Just skip it to maintain current behavior; in the future, we
            # may want to show an entry with some annotation to tell the
            # user about this sublayer.
            numMissingLayers = numMissingLayers + 1
        else:
            # Otherwise we expect the unmuted sublayer to be the next child
            # tree in the layer stack tree so we recursively add it.
            layers.extend(_AddLayerTree(
                stage, next(childTreeIter), depth=depth + 1))

    # Since we're relying on the correspondence between the unmuted sublayer 
    # paths and the child layer stack trees, report an error if the total number
    # of muted layers and child trees don't match up so we can track if it 
    # becomes an issue.
    if numMissingLayers + len(childTrees) != len(subLayerPaths):
        print("CODING ERROR: Encountered an unexpected number of muted "
              "sublayers of layer {}. The root layer stack may be "
              "incorrect in the layer stack view".format(
              layerTree.layer.identifier))

    return layers

def GetRootLayerStackInfo(stage):
    primIndex = stage.GetPseudoRoot().GetPrimIndex()
    layerStack = primIndex.rootNode.layerStack

    with Ar.ResolverContextBinder(stage.GetPathResolverContext()):
        return _AddLayerTree(stage, layerStack.layerTree)

def PrettyFormatSize(sz):
    k = 1024
    meg = k * 1024
    gig = meg * 1024
    ter = gig * 1024

    sz = float(sz)
    if sz > ter:
        return "%.1fT" % (sz/float(ter))
    elif sz > gig:
        return "%.1fG" % (sz/float(gig))
    elif sz > meg:
        return "%.1fM" % (sz/float(meg))
    elif sz > k:
        return "%.1fK" % (sz/float(k))
    else:
        return "%db" % sz


class Timer(object):
    """Use as a context object with python's "with" statement, like so:
       with Timer("do some stuff", printTiming=True):
           doSomeStuff()

       If you want to defer printing timing information, one way to do so is as
       follows:
       with Timer("do some stuff") as t:
           doSomeStuff()
       if wantToPrintTime:
           t.PrintTime()
    """

    # Annotate for performance tools if we're in the Pixar environment.
    # Silently skip this if the IttUtil module is not available.
    try:
        from pixar import IttUtil
    except ImportError:
        IttUtil = None

    def __init__(self, label, printTiming=False):
        self._printTiming = printTiming
        self._label = label
        self._isValid = False

    def __enter__(self):
        Trace.Collector().BeginEvent(self._label)
        self._stopwatch = Tf.Stopwatch()
        self._stopwatch.Start()
        self._isValid = True
        self.interval = 0
        if Timer.IttUtil:
            Timer.IttUtil.TaskBegin(self._label)
        return self

    def __exit__(self, excType, excVal, excTB):
        Trace.Collector().EndEvent(self._label)
        self._stopwatch.Stop()
        self.interval = self._stopwatch.seconds
        if Timer.IttUtil:
            Timer.IttUtil.TaskEnd()

        # Only report if we are valid and exiting cleanly (i.e. no exception).
        if self._printTiming and excType is None:
            self.PrintTime()

    def Invalidate(self):
        self._isValid = False

    def PrintTime(self):
        if self._isValid:
            print("Time to %s: %2.6fs" % (self._label, self.interval))

class BusyContext(object):
    """When used as a context object with python's "with" statement,
    will set Qt's busy cursor upon entry and pop it on exit.
    """
    def __enter__(self):
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.BusyCursor)

    def __exit__(self, *args):
        QtWidgets.QApplication.restoreOverrideCursor()


def InvisRootPrims(stage):
    """Make all defined root prims of stage be invisible,
    at Usd.TimeCode.Default()"""
    for p in stage.GetPseudoRoot().GetChildren():
        UsdGeom.Imageable(p).MakeInvisible()

def _RemoveVisibilityRecursive(primSpec):
    try:
        primSpec.RemoveProperty(primSpec.attributes[UsdGeom.Tokens.visibility])
    except IndexError:
        pass
    for child in primSpec.nameChildren:
        _RemoveVisibilityRecursive(child)

def ResetSessionVisibility(stage):
    session = stage.GetSessionLayer()
    with Sdf.ChangeBlock():
        _RemoveVisibilityRecursive(session.pseudoRoot)

# This is unfortunate.. but until UsdAttribute will return you a ResolveInfo,
# we have little alternative, other than manually walking prim's PcpPrimIndex
def HasSessionVis(prim):
    """Is there a session-layer override for visibility for 'prim'?"""
    session = prim.GetStage().GetSessionLayer()
    primSpec = session.GetPrimAtPath(prim.GetPath())
    return bool(primSpec and UsdGeom.Tokens.visibility in primSpec.attributes)

# This should be codified on UsdModelAPI, but maybe after 117137 is addressed...
def GetEnclosingModelPrim(prim):
    """If 'prim' is inside/under a model of any kind, return the closest
    such ancestor prim - If 'prim' has no model ancestor, return None"""
    if prim:
        prim = prim.GetParent()
    while prim:
        # We use Kind here instead of prim.IsModel because point instancer
        # prototypes currently don't register as models in IsModel. See
        # bug: http://bugzilla.pixar.com/show_bug.cgi?id=117137
        if Kind.Registry.IsA(Usd.ModelAPI(prim).GetKind(), Kind.Tokens.model):
            break
        prim = prim.GetParent()

    return prim

def GetPrimLoadability(prim):
    """Return a tuple of (isLoadable, isLoaded) for 'prim', according to
    the following rules:
    A prim is loadable if it is active, and either of the following are true:
       * prim has a payload
       * prim is a model group
    The latter is useful because loading is recursive on a UsdStage, and it
    is convenient to be able to (e.g.) load everything loadable in a set.

    A prim 'isLoaded' only if there are no unloaded prims beneath it, i.e.
    it is stating whether the prim is "fully loaded".  This
    is a debatable definition, but seems useful for usdview's purposes."""
    if not (prim.IsActive() and (prim.IsGroup() or prim.HasAuthoredPayloads())):
        return (False, True)
    # XXX Note that we are potentially traversing the entire stage here.
    # If this becomes a performance issue, we can cast this query into C++,
    # cache results, etc.
    for p in Usd.PrimRange(prim, Usd.PrimIsActive):
        if not p.IsLoaded():
            return (True, False)
    return (True, True)

def GetPrimsLoadability(prims):
    """Follow the logic of GetPrimLoadability for each prim, combining
    results so that isLoadable is the disjunction of all prims, and
    isLoaded is the conjunction."""
    isLoadable = False
    isLoaded = True
    for prim in prims:
        loadable, loaded = GetPrimLoadability(prim)
        isLoadable = isLoadable or loadable
        isLoaded = isLoaded and loaded
    return (isLoadable, isLoaded)

def GetFileOwner(path):
    try:
        if platform.system() == 'Windows':
            # This only works if pywin32 is installed.
            # Try "pip install pypiwin32".
            import win32security as w32s
            fs = w32s.GetFileSecurity(path, w32s.OWNER_SECURITY_INFORMATION)
            sdo = fs.GetSecurityDescriptorOwner()
            name, domain, use = w32.LookupAccountSid(None, sdo)
            return "%s\\%s" % (domain, name)
        else:
            import pwd
            return pwd.getpwuid(os.stat(path).st_uid).pw_name
    except:
        return "<unknown>"

# In future when we have better introspection abilities in Usd core API,
# we will change this function to accept a prim rather than a primStack.
def GetAssetCreationTime(primStack, assetIdentifier):
    """Finds the weakest layer in which assetInfo.identifier is set to
    'assetIdentifier', and considers that an "asset-defining layer".
    If assetInfo.identifier is not set in any layer, assumes the weakest
    layer is the defining layer.  We then retrieve the creation time for
    the asset by stat'ing the defining layer's real path.

    Returns a triple of strings: (fileDisplayName, creationTime, owner)"""
    definingLayer = None
    for spec in reversed(primStack):
        if spec.HasInfo('assetInfo'):
            identifier = spec.GetInfo('assetInfo').get('identifier')
            if identifier and identifier.path == assetIdentifier.path:
                definingLayer = spec.layer
                break
    if definingLayer:
        definingFile = definingLayer.realPath
    else:
        definingFile = primStack[-1].layer.realPath

    if Ar.IsPackageRelativePath(definingFile):
        definingFile = Ar.SplitPackageRelativePathOuter(definingFile)[0]

    if not definingFile:
        displayName = (definingLayer.GetDisplayName()
                       if definingLayer and definingLayer.anonymous else
                       "<in-memory layer>")
        creationTime = "<unknown>"
        owner = "<unknown>"
    else:
        displayName = definingFile.split('/')[-1]

        try:
            creationTime = time.ctime(os.stat(definingFile).st_ctime)
        except:
            creationTime = "<unknown>"

        owner = GetFileOwner(definingFile)

    return (displayName, creationTime, owner)


def DumpMallocTags(stage, contextStr):
    if Tf.MallocTag.IsInitialized():
        callTree = Tf.MallocTag.GetCallTree()
        memInMb = Tf.MallocTag.GetTotalBytes() / (1024.0 * 1024.0)

        import os.path as path
        import tempfile
        layerName = path.basename(stage.GetRootLayer().identifier)
        # CallTree.Report() gives us the most informative (and processable)
        # form of output, but it only accepts a fileName argument.  So we
        # use NamedTemporaryFile just to get a filename.
        statsFile = tempfile.NamedTemporaryFile(prefix=layerName+'.',
                                                suffix='.mallocTag',
                                                delete=False)
        statsFile.close()
        reportName = statsFile.name
        callTree.Report(reportName)
        print("Memory consumption of %s for %s is %d Mb" % (contextStr,
                                                            layerName,
                                                            memInMb))
        print("For detailed analysis, see " + reportName)
    else:
        print("Unable to accumulate memory usage since the Pxr MallocTag system was not initialized")

def GetInstanceIdForIndex(prim, instanceIndex, time):
    '''Attempt to find an authored Id value for the instance at index
    'instanceIndex' at time 'time', on the given prim 'prim', which we access
    as a UsdGeom.PointInstancer (whether it actually is or not, to provide
    some dynamic duck-typing for custom instancer types that support Ids.
    Returns 'None' if no ids attribute was found, or if instanceIndex is
    outside the bounds of the ids array.'''
    if not prim or instanceIndex < 0:
        return None
    ids = UsdGeom.PointInstancer(prim).GetIdsAttr().Get(time)
    if not ids or instanceIndex >= len(ids):
        return None
    return ids[instanceIndex]

def GetInstanceIndicesForIds(prim, instanceIds, time):
    '''Attempt to find the instance indices of a list of authored instance IDs
    for prim 'prim' at time 'time'. If the prim is not a PointInstancer or does
    not have authored IDs, returns None. If any ID from 'instanceIds' does not
    exist at the given time, its index is not added to the list (because it does
    not have an index).'''
    ids = UsdGeom.PointInstancer(prim).GetIdsAttr().Get(time)
    if ids:
        return [instanceIndex for instanceIndex, instanceId in enumerate(ids)
            if instanceId in instanceIds]
    else:
        return None

def Drange(start, stop, step):
    '''Return a list whose first element is 'start' and the following elements
    (if any) are 'start' plus increasing whole multiples of 'step', up to but
    not greater than 'stop'.  For example:
    Drange(1, 3, 0.3) -> [1, 1.3, 1.6, 1.9, 2.2, 2.5, 2.8]'''
    lst = [start]
    n = 1
    while start + n * step <= stop:
        lst.append(start + n * step)
        n += 1
    return lst

class PrimNotFoundException(Exception):
    """Raised when a prim does not exist at a valid path."""
    def __init__(self, path):
        super(PrimNotFoundException, self).__init__(
            "Prim not found at path in stage: %s" % str(path))

class PropertyNotFoundException(Exception):
    """Raised when a property does not exist at a valid path."""
    def __init__(self, path):
        super(PropertyNotFoundException, self).__init__(
            "Property not found at path in stage: %s" % str(path))

class FixableDoubleValidator(QtGui.QDoubleValidator):
    """This class implements a fixup() method for QDoubleValidator
    (see method for specific behavior).  To work around the brokenness
    of Pyside's fixup() wrapping, we allow the validator to directly
    update its parent if it is a QLineEdit, from within fixup().  Thus
    every QLineEdit must possess its own unique FixableDoubleValidator.
    
    The fixup method we supply (which can be usefully called directly)
    applies clamping and rounding to enforce the QDoubleValidator's
    range and decimals settings."""

    def __init__(self, parent):
        super(FixableDoubleValidator, self).__init__(parent)
        
        self._lineEdit = parent if isinstance(parent, QtWidgets.QLineEdit) else None

    def fixup(self, valStr):
        # We implement this to fulfill the virtual for internal QLineEdit
        # behavior, hoping that PySide will do the right thing, but it is
        # useless to call from Python directly due to string immutability
        try:
            val = float(valStr)
            val = max(val, self.bottom())
            val = min(val, self.top())
            val = round(val)
            valStr = str(val)
            if self._lineEdit:
                self._lineEdit.setText(valStr)
        except ValueError:
            pass



#
# Copyright 2021 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtCore, QtGui, QtWidgets
from .adjustFreeCameraUI import Ui_AdjustFreeCamera
from .common import FixableDoubleValidator

from pxr import Gf, UsdGeom

class AdjustFreeCamera(QtWidgets.QDialog):

    """Dialog to adjust the free camera settings (clipping, fov, aspect ratio).
    """
    def __init__(self, parent, dataModel, signalFrustumChanged):
        QtWidgets.QDialog.__init__(self,parent)
        self._ui = Ui_AdjustFreeCamera()
        self._ui.setupUi(self)

        self._dataModel = dataModel
        self._signalFrustumChanged = signalFrustumChanged

        self._signalFrustumChanged.connect(self._frustumChanged)

        # When the checkboxes change, we want to update instantly
        self._ui.overrideNear.stateChanged.connect(self._overrideNearToggled)

        self._ui.overrideFar.stateChanged.connect(self._overrideFarToggled)

        # we also want to update the clipping planes as the user is typing
        self._ui.nearSpinBox.valueChanged.connect(self._nearChanged)

        self._ui.farSpinBox.valueChanged.connect(self._farChanged)

        # Set the checkboxes to their initial state
        self._ui.overrideNear.setChecked(
            self._dataModel.viewSettings.freeCameraOverrideNear is not None)
        self._ui.overrideFar.setChecked(
            self._dataModel.viewSettings.freeCameraOverrideFar is not None)

        # load the initial values for the text boxes, but first deactivate them
        # if their corresponding checkbox is off.
        clipRange = self._getCurrentClippingRange()
        self._ui.nearSpinBox.setEnabled(self._ui.overrideNear.isChecked())
        self._ui.nearSpinBox.setValue(clipRange.min)
        self._ui.farSpinBox.setEnabled(self._ui.overrideFar.isChecked())
        self._ui.farSpinBox.setValue(clipRange.max)

        self._ui.lockFreeCamAspect.setChecked(
            bool(self._dataModel.viewSettings.lockFreeCameraAspect))
        self._ui.freeCamAspect.setValue(self._getCurrentAspectRatio())
        self._ui.freeCamAspect.setEnabled(
            bool(self._dataModel.viewSettings.lockFreeCameraAspect))

        self._ui.lockFreeCamAspect.stateChanged.connect(
            self._lockFreeCamAspectToggled)
        self._ui.freeCamAspect.valueChanged.connect(self._aspectSpinBoxChanged)

        self._ui.freeCamFov.setValue(
            self._dataModel.viewSettings.freeCameraFOV)
        self._ui.freeCamFov.valueChanged.connect(self._freeCamFovChanged)

    def _overrideNearToggled(self, state):
        """Called when the "Override Near" checkbox is toggled"""
        self._ui.nearSpinBox.setEnabled(state)
        if state:
            self._dataModel.viewSettings.freeCameraOverrideNear = \
                self._getCurrentClippingRange().min
        else:
            self._dataModel.viewSettings.freeCameraOverrideNear = None

    def _overrideFarToggled(self, state):
        """Called when the "Override Far" checkbox is toggled"""
        self._ui.farSpinBox.setEnabled(state)
        if state:
            self._dataModel.viewSettings.freeCameraOverrideFar = \
                self._getCurrentClippingRange().max
        else:
            self._dataModel.viewSettings.freeCameraOverrideFar = None

    def _nearChanged(self, value):
        """Called when the Near spin box changed.  This can happen when we
        are updating the value but the widget is actually inactive - don't
        do anything in that case."""
        if self._ui.nearSpinBox.isEnabled():
            self._dataModel.viewSettings.freeCameraOverrideNear = value

    def _farChanged(self, value):
        """Called when the Far spin box changed.  This can happen when we
        are updating the value but the widget is actually inactive - don't
        do anything in that case."""
        if self._ui.farSpinBox.isEnabled():
            self._dataModel.viewSettings.freeCameraOverrideFar = value

    def _lockFreeCamAspectToggled(self, state):
        lockFreeCameraAspect = QtCore.Qt.CheckState(state) != QtCore.Qt.CheckState.Unchecked
        self._ui.freeCamAspect.setEnabled(lockFreeCameraAspect)
        self._dataModel.viewSettings.lockFreeCameraAspect = \
            lockFreeCameraAspect
        if lockFreeCameraAspect:
            self._dataModel.viewSettings.freeCameraAspect = \
                self._ui.freeCamAspect.value()

    def _aspectSpinBoxChanged(self, value):
        """Updates the camera's aspect ratio based on the spin box value."""
        if self._dataModel.viewSettings.lockFreeCameraAspect:
            self._dataModel.viewSettings.freeCameraAspect = value

    def _getCurrentAspectRatio(self):
        """Returns the current aspect ratio that should be displayed in the spin
        box.

        If a camera prim is active, reflect that value. Otherwise, use the
        current setting."""
        cameraPrim = self._dataModel.viewSettings.cameraPrim
        if cameraPrim and cameraPrim.IsActive():
            gfCam = UsdGeom.Camera(cameraPrim).GetCamera(
                self._dataModel.currentFrame)
            return gfCam.aspectRatio

        return self._dataModel.viewSettings.freeCameraAspect

    def _getCurrentFov(self):
        """Returns the current vertical field of view that should be displayed
        in the spin box.

        If a camera prim is active, reflect that value. Otherwise, use the
        current setting."""
        cameraPrim = self._dataModel.viewSettings.cameraPrim
        if cameraPrim and cameraPrim.IsActive():
            gfCam = UsdGeom.Camera(cameraPrim).GetCamera(
                self._dataModel.currentFrame)
            return gfCam.GetFieldOfView(Gf.Camera.FOVVertical)

        return self._dataModel.viewSettings.freeCameraFOV

    def _getCurrentClippingRange(self):
        """Returns the current clipping range (near, far) that should be
        displayed in the spin boxes.

        If the view settings have values for freeCameraOverrideNear/Far, then
        those values should be chosen. Otherwise, take the clipping range from
        the current camera (whether a camera prim or the free camera)."""

        viewSettings = self._dataModel.viewSettings
        near = viewSettings.freeCameraOverrideNear
        far = viewSettings.freeCameraOverrideFar
        if near is not None and far is not None:
            return Gf.Range1f(near, far)

        cameraPrim = viewSettings.cameraPrim
        if cameraPrim and cameraPrim.IsActive():
            gfCam = UsdGeom.Camera(cameraPrim).GetCamera(
                self._dataModel.currentFrame)
            if near is None:
                near = gfCam.clippingRange.min
            if far is None:
                far = gfCam.clippingRange.max
        elif viewSettings.freeCamera:
            if near is None:
                near = viewSettings.freeCamera.near
            if far is None:
                far = viewSettings.freeCamera.far

        if near is None:
            near = 1.0
        if far is None:
            far = 1e6
        return Gf.Range1f(near, far)

    def _freeCamFovChanged(self, value):
        self._dataModel.viewSettings.freeCameraFOV = value

    def _frustumChanged(self):
        """Updates the UI to reflect the current camera frustum."""
        self._ui.freeCamAspect.blockSignals(True)
        self._ui.freeCamAspect.setValue(self._getCurrentAspectRatio())
        self._ui.freeCamAspect.blockSignals(False)

        self._ui.freeCamFov.blockSignals(True)
        self._ui.freeCamFov.setValue(self._getCurrentFov())
        self._ui.freeCamFov.blockSignals(False)

        clippingRange = self._getCurrentClippingRange()
        self._ui.nearSpinBox.blockSignals(True)
        self._ui.nearSpinBox.setValue(clippingRange.min)
        self._ui.nearSpinBox.blockSignals(False)
        self._ui.farSpinBox.blockSignals(True)
        self._ui.farSpinBox.setValue(clippingRange.max)
        self._ui.farSpinBox.blockSignals(False)

        self.update()

    def closeEvent(self, event):
        # Ensure that even if the dialog doesn't get destroyed right away,
        # we'll stop doing work.
        self._signalFrustumChanged.disconnect(self._frustumChanged)

        event.accept()
        # Since the dialog is the immediate-edit kind, we consider
        # window-close to be an accept, so our clients can know the dialog is
        # done by listening for the finished(int) signal
        self.accept()
